import numpy as np

from .base import Logical
from .result import Vector, InfiniteVector, join
from .results import Results


class ProbabilitySpace:
    """Defines a probability space.

    Attributes:
      draw (function): A function explaining how to draw one
        outcome from the probability space.
    """

    def __init__(self, draw):
        self.draw = draw

    def sim(self, n):
        """Simulate n draws from probability space.

        Args:
          n (int): How many draws to make.

        Returns:
          Results: A list-like object containing the simulation results.
        """
        return Results(self.draw() for _ in range(n))

    def check_same(self, other):
        if self != other:
            raise Exception("Events must be defined on same probability space.")

    def apply(self, func):
        """Define a new probability space.

        Args:
          func: function to apply to each realization.

        Returns:
          A new ProbabilitySpace where each realization is func applied to
          a realization from the current probability space.
        """
        def draw():
            return func(self.draw())
        return ProbabilitySpace(draw)

    def __mul__(self, other):
        def draw():
            return join(self.draw(), other.draw())
        return ProbabilitySpace(draw)

    def __pow__(self, exponent):
        if exponent == float("inf"):
            def draw():
                def _func(_):
                    return self.draw()
                return InfiniteVector(_func)
        else:
            def draw():
                return Vector(self.draw() for _ in range(exponent))
        return ProbabilitySpace(draw)


class Event(Logical):

    def __init__(self, prob_space, func):
        self.prob_space = prob_space
        self.func = func

    def check_same_prob_space(self, other):
        self.prob_space.check_same(other.prob_space)

    # The Logical superclass will use this to define the three
    # logical operations: and (&), or (|), not (~).
    def _logical_factory(self, op):

        def _op_func(self, other=None):
            # other will be None when op is the "not" operator
            if other is None:
                return Event(self.prob_space,
                             lambda outcome: op(self.func(outcome)))
            else:
                if isinstance(other, Event):
                    self.check_same_prob_space(other)
                else:
                    raise TypeError(
                        "Logical operations are only defined "
                        "between two Events, not between an Event "
                        "and a %s." % type(other).__name__)
                return Event(self.prob_space,
                             lambda outcome: op(self.func(outcome),
                                                other.func(outcome)))

        return _op_func

    # This prevents users from writing expressions like 2 < X < 5,
    # which evaluate to ((2 < X) and (X < 5)). This unfortunately
    # is not well-defined in Python and cannot be overloaded.
    def __bool__(self):
        raise Exception("Cannot cast an Event to a boolean. "
                        "You may be getting this error if you "
                        "wrote an expression like (2 < X < 5). "
                        "Try ((2 < X) & (X < 5)) instead.")

    def draw(self):
        return self.func(self.prob_space.draw())

    def sim(self, n):
        return Results(self.draw() for _ in range(n))


class BoxModel(ProbabilitySpace):
    """Defines a probability space from a box model.

    Attributes:
      box (list-like or dict-like): The box to sample from.
        The box can be specified either directly as a list
        of objects or indirectly as a dict of objects and
        their counts.
      size (int): How many draws to make.
      replace (bool-like): Sample with replacement or without?
      probs (list): Probabilities of sampling each ticket
        (by default, all tickets are equally likely). Note
        that this is ignored if box is specified as a dict.
      order_matters (bool): Should we count different
        orderings of the same tickets as different outcomes?
        Essentially, this determines whether the draws are
        sorted before returning or not.
    """

    def __init__(self, box, size=None, replace=True, probs=None, order_matters=True):
        if isinstance(box, list):
            self.box = box
            self.probs = probs
        elif isinstance(box, dict):
            self.box = []
            for ticket, count in box.items():
                self.box.extend([ticket] * count)
            self.probs = None
        else:
            raise Exception(
                "Box must be specified either as a list or a dict."
            )
        self.size = None if size == 1 else size
        self.replace = replace
        self.order_matters = order_matters
        self.output_type = Vector
        self.infinite_output_type = InfiniteVector

        # If drawing without replacement, check that the number
        # of draws does not exceed the number of tickets in the box.
        if not self.replace and self.size > len(self.box):
            raise Exception(
                "Cannot draw more tickets (without replacement) "
                "than there are tickets in the box."
            )

    def draw(self):
        """
        A function that takes no arguments and returns a value(s) from the
            "box" argument of the BoxModel.

        Based on BoxModel inputs:
        Number of values returned depends on the input of the "size"
            argument.
        Whether or not a value in the box can appear multiple times
            depends on the "replace" argument.
        If a list of probabilities is specified, values drawn will be drawn
            with the specified probabilities.
        """

        def draw_inds(size):
            return np.random.choice(len(self.box), size, self.replace, self.probs)

        if self.size is None:
            return self.box[draw_inds(None)]
        elif self.size == float("inf"):
            def _func(_):
                return self.box[draw_inds(None)]
            return self.infinite_output_type(_func)
        else:
            draws = [self.box[i] for i in draw_inds(self.size)]
            if not self.order_matters:
                draws.sort()
            return self.output_type(draws)


class DeckOfCards(BoxModel):
    """Defines the probability space for drawing from a deck of cards.

    Attributes:
      size (int): How many draws to make.
      replace (bool): Sample with replacement or without?
      order_matters (bool): Should we count different orderings of
        the same cards as different outcomes or the same outcome?
    """

    def __init__(self, size=None, replace=False, order_matters=True):
        box = []
        for rank in list(range(2, 11)) + ["J", "Q", "K", "A"]:
            for suit in ["Diamonds", "Hearts", "Clubs", "Spades"]:
                box.append((rank, suit))
        super().__init__(box, size, replace,
                         probs=None, order_matters=order_matters)



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde

figure = plt.figure

xlabel = plt.xlabel
ylabel = plt.ylabel

xlim = plt.xlim
ylim = plt.ylim

def get_next_color(axes):
    color_cycle = axes._get_lines.prop_cycler
    color = next(color_cycle)["color"]
    return color

def configure_axes(axes, xdata, ydata, xlabel = None, ylabel = None):
    # Create 5% buffer on either end of plot so that leftmost and rightmost
    # lines are visible. However, if current axes are already bigger,
    # keep current axes.
    buff = .05 * (max(xdata) - min(xdata))
    xmin, xmax = axes.get_xlim()
    xmin = min(xmin, min(xdata) - buff)
    xmax = max(xmax, max(xdata) + buff)
    plt.xlim(xmin, xmax)

    _, ymax = axes.get_ylim()
    ymax = max(ymax, 1.05 * max(ydata))
    plt.ylim(0, ymax)
    
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

def plot(*args, **kwargs):
    try:
        args[0].plot(**kwargs)
    except:
        plt.plot(*args, **kwargs)
        
def is_discrete(heights):
    return sum([(i > 1) for i in heights]) > .8 * len(heights)

def count_var(x):
    counts = {}
    for val in x:
        if val in counts:
            counts[val] += 1
        else:
            counts[val] = 1
    return counts
    
def compute_density(values):
    density = gaussian_kde(values)
    density.covariance_factor = lambda: 0.25
    density._compute_covariance()
    return density

def setup_ticks(pos, lab, ax):
    ax.set_ticks(pos)
    ax.set_ticklabels(lab)
    
def add_colorbar(fig, type, mappable, label):
    #create axis for cbar to place on left
    if 'marginal' not in type: 
        caxes = fig.add_axes([0, 0.1, 0.05, 0.8])
    else: #adjust height if marginals
        caxes = fig.add_axes([0, 0.1, 0.05, 0.57])
    cbar = plt.colorbar(mappable=mappable, cax=caxes)
    caxes.yaxis.set_ticks_position('left')
    cbar.set_label(label)
    caxes.yaxis.set_label_position('left')
    return caxes

def setup_tile(v, bins, discrete):
    if not discrete:
        v_lab = np.linspace(min(v), max(v), bins + 1)
        v_pos = np.arange(0, len(v_lab)) - 0.5
        v_vect = np.digitize(v, v_lab, right=True) - 1
    else:
        v_lab = np.unique(v) #returns sorted array
        v_pos = range(len(v_lab))
        v_map = dict(zip(v_lab, v_pos))
        v_vect = np.vectorize(v_map.get)(v)
    return v_vect, v_lab, v_pos

def make_tile(x, y, bins, discrete_x, discrete_y, ax):
    x_vect, x_lab, x_pos = setup_tile(x, bins, discrete_x)
    y_vect, y_lab, y_pos = setup_tile(y, bins, discrete_y)
    nums = len(x_vect)
    counts = count_var(list(zip(y_vect, x_vect)))
    y_shape = len(y_lab) if discrete_y else len(y_lab) - 1
    x_shape = len(x_lab) if discrete_x else len(x_lab) - 1
    intensity = np.zeros(shape=(y_shape, x_shape))
        
    for key, val in counts.items():
        intensity[key] = val / nums
    if not discrete_x: x_lab = np.around(x_lab, decimals=1)
    if not discrete_y: y_lab = np.around(y_lab, decimals=1)
    hm = ax.matshow(intensity, cmap='Blues', origin='lower', aspect='auto', vmin=0)
    ax.xaxis.set_ticks_position('bottom')
    setup_ticks(x_pos, x_lab, ax.xaxis)
    setup_ticks(y_pos, y_lab, ax.yaxis)
    return hm

def make_violin(data, positions, ax, axis, alpha):
    values = []
    i, j = (0, 1) if axis == 'x' else (1, 0)
    values = [data[data[:, i] == pos, j].tolist() for pos in positions]
    violins = ax.violinplot(dataset=values, showmedians=True,
                            vert=False if axis == 'y' else True)
    setup_ticks(np.array(positions) + 1, positions, 
                ax.xaxis if axis == 'x' else ax.yaxis)
    for part in violins['bodies']:
        part.set_edgecolor('black')
        part.set_alpha(alpha)
    for component in ('cbars', 'cmins', 'cmaxes', 'cmedians'):
        vp = violins[component]
        vp.set_edgecolor('black')
        vp.set_linewidth(1)

def make_marginal_impulse(count, color, ax_marg, alpha, axis):
    key, val = list(count.keys()), list(count.values())
    tot = sum(val)
    val = [i / tot for i in val]
    if axis == 'x':
        ax_marg.vlines(key, 0, val, color=color, alpha=alpha)
    elif axis == 'y':
        ax_marg.hlines(key, 0, val, color=color, alpha=alpha)

def make_density2D(x, y, ax):
    res = np.vstack([x, y])
    density = gaussian_kde(res)
    xmax, xmin = max(x), min(x)
    ymax, ymin = max(y), min(y)
    Xgrid, Ygrid = np.meshgrid(np.linspace(xmin, xmax, 100),
                               np.linspace(ymin, ymax, 100))
    Z = density.evaluate(np.vstack([Xgrid.ravel(), Ygrid.ravel()]))
    den = ax.imshow(Z.reshape(Xgrid.shape), origin='lower', cmap='Blues',
              aspect='auto', extent=[xmin, xmax, ymin, ymax]
    )
    return den



import math
import numbers
import operator as op

import numpy as np
import scipy.stats as stats

from .random_variables import RV
from .result import (
    Tuple,
    TimeFunction,
    ContinuousTimeFunction,
    DiscreteValued
)
from .results import Results

pi = math.pi
e = math.e
inf = float("inf")

floor = math.floor
ceil = math.ceil

def operation_factory(operation):

    def _op_func(x):
        if isinstance(x, (RV, Tuple, TimeFunction)):
            # recursively call op_fun until x is a scalar
            return x.apply(_op_func)
        elif isinstance(x, Results):
            return x.apply(_op_func)
        else:
            return operation(x)

    return _op_func

sqrt = operation_factory(math.sqrt)
exp = operation_factory(math.exp)
sin = operation_factory(math.sin)
cos = operation_factory(math.cos)
tan = operation_factory(math.tan)
factorial = operation_factory(math.factorial)

def log(value, base=e):
    return operation_factory(lambda x: math.log(x, base))(value)

def mean(x):
    if isinstance(x, numbers.Real):
        raise Exception("Taking the mean with one value is unnecessary.")
    else:
        return sum(x) / len(x)

def cumsum(x):
    return x.cumsum()

def var(x):
    return mean([(i - mean(x)) ** 2 for i in x])

def sd(x):
    return math.sqrt(var(x))

def median(x):
    if isinstance(x, numbers.Real):
        raise Exception("Taking the median of one value is unnecessary.")
    else:
        return np.median(x)

def min_max_diff(x):
    if isinstance(x, numbers.Real):
        raise Exception("Taking the range of one value is unnecessary.")
    else:
        return max(x) - min(x)

def med_abs_dev(x):
    return median(list(abs(i-median(x)) for i in x))

def quantile(q):
    return lambda x: np.percentile(x, q * 100)

def iqr(x):
    if isinstance(x, numbers.Real):
        raise Exception("Taking the iqr of one value is unnecessary.")
    else:
        q75, q25 = np.percentile(x, [75, 25])
        return q75 - q25

def orderstatistics(n):
    if n <= 0:
        raise Exception("Out of bounds. Lowest order is 1.")
    else:
        return lambda x: np.partition(x, n - 1)[n - 1]

def skewness(x):
    if isinstance(x, numbers.Real):
        raise Exception("Finding the skenewss of one value is unnecessary,")
    else:
        return stats.skew(x)

def kurtosis(x):
    if isinstance(x, numbers.Real):
        raise Exception("Finding the kurtosis of one value is unnecessary.")
    else:
        return stats.kurtosis(x)

def moment(k):
    return lambda x: stats.moment(x, k)

def trimmed_mean(alpha):
    return lambda x: stats.trim_mean(x, alpha)

def comparefun(x, compare, value):
    count = 0
    for i in x:
        if compare(i, value):
            count += 1
    return count

def count(func=lambda x: True):
    def _func(x):
        val = 0
        for i in x:
            if func(i):
                val += 1
        return val
    return _func

def count_eq(value):
    def func(x):
        return comparefun(x, op.eq, value)
    return func

def count_neq(value):
    def func(x):
        return comparefun(x, op.ne, value)
    return func

def count_lt(value):
    def func(x):
        return comparefun(x, op.lt, value)
    return func

def count_gt(value):
    def func(x):
        return comparefun(x, op.gt, value)
    return func

def count_geq(value):
    def func(x):
        return comparefun(x, op.ge, value)
    return func

def count_leq(value):
    def func(x):
        return comparefun(x, op.le, value)
    return func

def interarrival_times(continuous_time_function):
    """Given a realization of a continuous-time,
       discrete-state process, returns the interarrival
       times (i.e., the times between each state change).

    Args:
      continuous_time_function: A ContinuousTimeFunction
        object, such as ContinuousTimeMarkovChainResult or
        PoissonProcessResult.
    """
    if not (isinstance(continuous_time_function,
                       ContinuousTimeFunction) and
            isinstance(continuous_time_function,
                       DiscreteValued)):
        raise TypeError(
            "Interarrival times are only defined for "
            "continuous-time, discrete-valued functions."
        )
    return continuous_time_function.get_interarrival_times()

def arrival_times(continuous_time_function):
    """Given a realization of a continuous-time,
       discrete-state process, returns the arrival
       times (i.e., the times when the state changes).

    Args:
      continuous_time_function: A ContinuousTimeFunction
        object, such as ContinuousTimeMarkovChainResult or
        PoissonProcessResult.
    """
    if not (isinstance(continuous_time_function,
                       ContinuousTimeFunction) and
            isinstance(continuous_time_function,
                       DiscreteValued)):
        raise TypeError(
            "Interarrival times are only defined for "
            "continuous-time, discrete-valued functions."
        )
    return continuous_time_function.get_arrival_times()

def states(discrete_valued_function):
    """Given a realization of a discrete-valued function,
       returns an InfiniteVector of the sequence of
       values (or states).

    Args:
      discrete_valued_function: A DiscreteValued object
                                (e.g., MarkovChainResult or
                                       PoissonProcessResult)
    """
    if not isinstance(discrete_valued_function, DiscreteValued):
        raise TypeError(
            "States are only defined for discrete-valued "
            "functions."
        )
    return discrete_valued_function.get_states()



import numpy as np

from .distributions import Exponential
from .math import inf
from .probability_space import ProbabilitySpace
from .random_variables import RV
from .result import (
    InfiniteVector, ContinuousTimeFunction, DiscreteValued
)

EPS = 1e-15


class MarkovChainResult(InfiniteVector, DiscreteValued):

    def __init__(self, transition_matrix, initial_dist, state_labels=None):
        # Check transition matrix
        for row in transition_matrix:
            if abs(sum(row) - 1) > EPS:
                raise Exception("Rows of a transition matrix must sum to 1.")
            for p in row:
                if p < 0:
                    raise Exception("Probabilities cannot be negative.")
        # Check that dimensions agree
        self.transition_matrix = np.asarray(transition_matrix)
        m, n = self.transition_matrix.shape
        if m != n:
            raise Exception("Transition matrix must be square.")
        if len(initial_dist) != n:
            raise Exception("Initial distribution must be a vector whose "
                            "length matches the dimensions of the "
                            "transition matrix.")
        self.initial_dist = initial_dist
        # Process state labels
        if state_labels is not None:
            if len(state_labels) != n:
                raise Exception("There must be as many state labels as "
                                "there are states.")
            self.state_labels = state_labels
        else:
            self.state_labels = range(n)
        self.n_states = n

        # Generate initial state.
        # (self.states stores the indexes of the states, while
        #  self.values stores the labels of the states.)
        state = np.random.choice(range(n), p=self.initial_dist)
        self.states = [state]

        def _func(n):
            m = len(self.states)
            # If nth state not generated yet, generate it.
            if n >= m:
                state = self.states[m - 1]
                for _ in range(m, n + 1):
                    state = np.random.choice(
                        range(self.n_states),
                        p=self.transition_matrix[state, :]
                    )
                    self.states.append(state)
            else:
                state = self.states[n]
            return self.state_labels[state]

        super().__init__(_func)

    def get_states(self):
        return self


class MarkovChainProbabilitySpace(ProbabilitySpace):

    def __init__(self, transition_matrix, initial_dist, state_labels=None):
        """Initialize probability space for a (discrete-time) Markov chain.

        Args:
          transition_matrix: n x n transition matrix
          initial_dist: length n vector of the initial distribution
          state_labels: length n vector of the labels of each state
                        (defaults to 0, 1, ..., n-1)
        """

        def _draw():
            return MarkovChainResult(transition_matrix,
                                     initial_dist,
                                     state_labels)

        super().__init__(_draw)


class MarkovChain(RV):

    def __init__(self, transition_matrix, initial_dist, state_labels=None):
        """Initialize a (discrete-time) Markov chain.

        Args:
          transition_matrix: n x n transition matrix
          initial_dist: length n vector of the initial distribution
          state_labels: length n vector of the labels of each state
                        (defaults to 0, 1, ..., n-1)
        """

        prob_space = MarkovChainProbabilitySpace(
            transition_matrix,
            initial_dist,
            state_labels)
        super().__init__(prob_space)


class ContinuousTimeMarkovChainResult(ContinuousTimeFunction,
                                      DiscreteValued):

    def __init__(self, states, rates,
                 unscaled_interarrival_times,
                 state_labels):
        self.states = states
        self.rates = rates
        self.times = unscaled_interarrival_times
        self.state_labels = state_labels

        # Define an InfiniteVector of the interarrival times.
        def interarrival_times(n):
            for i in range(n + 1):
                state = self.states[i]
                interarrival_time = self.times[i] / self.rates[state]
            return interarrival_time
        self.interarrival_times = InfiniteVector(interarrival_times)

        def _func(t):
            total_time = 0
            n = 0
            while True:
                state = self.states[n]
                total_time += self.times[n] / self.rates[state]
                if total_time > t:
                    return self.state_labels[state]
                n += 1
        super().__init__(_func)


class ContinuousTimeMarkovChainProbabilitySpace(ProbabilitySpace):

    def __init__(self, generator_matrix, initial_dist, state_labels=None):
        """Initialize a probability space for a continuous-time Markov chain.

        Args:
          generator_matrix: n x n generator matrix whose rows sum to 0
          initial_dist: length n vector of the initial distribution
          state_labels: length n vector of the labels of each state
                        (defaults to 0, 1, ..., n-1)
        """

        # Check generator matrix
        for i, row in enumerate(generator_matrix):
            if abs(sum(row)) > EPS:
                raise Exception("Rows of a generator matrix must sum to 0.")
            for j, q in enumerate(row):
                if j == i:
                    if q > 0:
                        raise Exception("Diagonal elements of a generator matrix " +
                                        "cannot be positive.")
                else:
                    if q < 0:
                        raise Exception("Off-diagonal elements of a generator matrix " +
                                        "cannot be negative.")
        # Check that dimensions agree
        self.generator_matrix = np.array(generator_matrix)
        m, n = self.generator_matrix.shape
        if m != n:
            raise Exception("Transition matrix must be square.")
        if len(initial_dist) != n:
            raise Exception("Initial distribution must be a vector whose "
                            "length matches the dimensions of the "
                            "transition matrix.")
        self.initial_dist = initial_dist
        # Process state labels
        if state_labels is not None:
            if len(state_labels) != n:
                raise Exception("There must be as many state labels as "
                                "there are states.")
            self.state_labels = state_labels
        else:
            self.state_labels = range(n)
        self.n_states = n

        # determine transition matrix
        transition_matrix = []
        for i, row in enumerate(self.generator_matrix):
            rate = -row[i]
            transition_matrix.append(
                [p / rate if j != i else 0 for j, p in enumerate(row)]
            )
        self.transition_matrix = np.array(transition_matrix)

        # A continuous-time Markov chain is specified by the
        # sequence of states and the unscaled interarrival times.
        def _draw():
            states = MarkovChain(self.transition_matrix,
                                 self.initial_dist).draw()
            rates = -np.diag(self.generator_matrix)
            unscaled_interarrival_times = (Exponential(1) ** inf).draw()
            return ContinuousTimeMarkovChainResult(
                states,
                rates,
                unscaled_interarrival_times,
                self.state_labels)

        super().__init__(_draw)


class ContinuousTimeMarkovChain(RV):

    def __init__(self, generator_matrix, initial_dist, state_labels=None):
        """Initialize a continuous-time Markov chain.

        Args:
          generator_matrix: n x n generator matrix whose rows sum to 0
          initial_dist: length n vector of the initial distribution
          state_labels: length n vector of the labels of each state
                        (defaults to 0, 1, ..., n-1)
        """

        prob_space = ContinuousTimeMarkovChainProbabilitySpace(
            generator_matrix,
            initial_dist,
            state_labels)
        super().__init__(prob_space)



import math

import numpy as np
import scipy.stats as stats

class Arithmetic:
    """A class with operations such as +, -, *, /.

    Subclasses must implement the _operation_factory method,
    which specifies how each operation acts on instances of
    that class.
    """

    # e.g., X + Y or X + 3
    def __add__(self, other):
        op_func = self._operation_factory(lambda x, y: x + y)
        return op_func(self, other)

    # e.g., 3 + X
    def __radd__(self, other):
        return self.__add__(other)

    # e.g., X - Y or X - 3
    def __sub__(self, other):
        op_func = self._operation_factory(lambda x, y: x - y)
        return op_func(self, other)

    # e.g., 3 - X
    def __rsub__(self, other):
        return -1 * self.__sub__(other)

    # e.g., -X
    def __neg__(self):
        return -1 * self

    # e.g., X * Y or X * 2
    def __mul__(self, other):
        op_func = self._operation_factory(lambda x, y: x * y)
        return op_func(self, other)

    # e.g., 2 * X
    def __rmul__(self, other):
        return self.__mul__(other)

    # e.g., X / Y or X / 2
    def __truediv__(self, other):
        op_func = self._operation_factory(lambda x, y: x / y)
        return op_func(self, other)

    # e.g., 2 / X
    def __rtruediv__(self, other):
        op_func = self._operation_factory(lambda x, y: y / x)
        return op_func(self, other)

    # e.g., X ** 2
    def __pow__(self, other):
        op_func = self._operation_factory(lambda x, y: x ** y)
        return op_func(self, other)

    # e.g., 2 ** X
    def __rpow__(self, other):
        op_func = self._operation_factory(lambda x, y: y ** x)
        return op_func(self, other)

    # Alternative notation for powers: e.g., X ^ 2
    def __xor__(self, other):
        return self.__pow__(other)

    # Alternative notation for powers: e.g., 2 ^ X
    def __rxor__(self, other):
        return self.__rpow__(other)


class Comparable:
    """A class with comparison operators such as <, >, and ==.

    Subclasses must implement the _comparison_factory method,
    which specifies how each comparison acts on instances of
    that class.
    """

    def __eq__(self, other):
        op_func = self._comparison_factory(lambda x, y: x == y)
        return op_func(self, other)

    def __ne__(self, other):
        op_func = self._comparison_factory(lambda x, y: x != y)
        return op_func(self, other)

    def __lt__(self, other):
        op_func = self._comparison_factory(lambda x, y: x < y)
        return op_func(self, other)

    def __le__(self, other):
        op_func = self._comparison_factory(lambda x, y: x <= y)
        return op_func(self, other)

    def __gt__(self, other):
        op_func = self._comparison_factory(lambda x, y: x > y)
        return op_func(self, other)

    def __ge__(self, other):
        op_func = self._comparison_factory(lambda x, y: x >= y)
        return op_func(self, other)


class Statistical:
    """A class with statistical functions, such as mean, var, etc.

    Subclasses must implement the _statistic_factory and 
    _multivariate_statistic_factory methods, which specify how
    (univariate) statistics (e.g., mean and variance), as well as
    multivariate statistics (e.g., covariance and correlation)
    are calculated on the object.
    """

    def sum(self):
        r"""Calculate the sum.

        .. math:: \frac{1}{n} \sum_{i=1}^n x_i

        Returns:
          The sum of the numbers.
        """
        op_func = self._statistic_factory(np.sum)
        return op_func(self)

    def mean(self):
        r"""Calculate the mean (a.k.a. average).

        The mean, or average, is a measure of center.

        .. math:: \mu = \frac{1}{n} \sum_{i=1}^n x_i

        Returns:
          float: The mean of the numbers.
        """
        op_func = self._statistic_factory(np.mean)
        return op_func(self)

    def quantile(self, q):
        r"""Calculate a specified quantile (percentile).

        The (100q)th quantile is the value x such that

        .. math:: \frac{\#\{ i: x_i \leq x \}}{n} = q

        Args:
          q (float): A number between 0 and 1 specifying
            the desired quantile or percentile.

        Returns:
          The (100q)th quantile of the numbers.
        """
        op_func = self._statistic_factory(
            lambda **kwargs: np.percentile(q=q * 100, **kwargs)
        )
        return op_func(self)

    def percentile(self, q):
        r"""Calculate a specified percentile.

        Alias for .quantile().
        """
        return self.quantile(q)

    def iqr(self):
        r"""Calculate the interquartile range (IQR).

        The IQR is the 75th percentile minus the 25th percentile.

        Returns:
          The interquartile range.
        """
        return self.quantile(.75) - self.quantile(.25)

    def median(self):
        r"""Calculate the median.

        The median is the middle number in a *sorted* list.
        It is a measure of center.

        Returns:
          The median of the numbers.
        """
        op_func = self._statistic_factory(np.median)
        return op_func(self)

    def std(self):
        r"""Calculate the standard deviation.

        The standard deviation is the square root of the variance.
        It is a measure of spread.

        .. math::

        \sigma &= \sqrt{\frac{1}{n} \sum_{i=1}^n (x_i - \mu)^2} \\
               &= \sqrt{\frac{1}{n} \sum_{i=1}^n x_i^2 - \mu^2}

        Returns:
          float: The standard deviation of the numbers.
        """
        op_func = self._statistic_factory(np.std)
        return op_func(self)

    def sd(self):
        r"""Calculate the standard deviation.

        The standard deviation is the square root of the variance.
        It is a measure of spread.

        .. math::

        \sigma &= \sqrt{\frac{1}{n} \sum_{i=1}^n (x_i - \mu)^2} \\
               &= \sqrt{\frac{1}{n} \sum_{i=1}^n x_i^2 - \mu^2}

        Returns:
          float: The standard deviation of the numbers.
        """
        return self.std()

    def var(self):
        r"""Calculate the variance.

        The variance is the average squared distance between
        each number and the mean. It is a measure of spread.

        .. math::

        \sigma^2 &= \frac{1}{n} \sum_{i=1}^n (x_i - \mu)^2 \\
               &= \frac{1}{n} \sum_{i=1}^n x_i^2 - \mu^2

        Returns:
          float: The variance of the numbers.
        """
        op_func = self._statistic_factory(np.var)
        return op_func(self)

    def skew(self):
        r"""Calculate the skewness.

        Returns:
          The skewness of the numbers.
        """
        op_func = self._statistic_factory(stats.skew)
        return op_func(self)

    def skewness(self):
        """Calculate the skewness. Alias for .skew()"""
        return self.skew()

    def kurtosis(self):
        r"""Calculate the kurtosis.

        Returns:
          The kurtosis of the numbers.
        """
        op_func = self._statistic_factory(stats.kurtosis)
        return op_func(self)

    def max(self):
        r"""Calculate the maximum.

        Returns:
          The maximum of the numbers.
        """
        op_func = self._statistic_factory(np.amax)
        return op_func(self)

    def min(self):
        r"""Calculate the minimum.

        Returns:
          The minimum of the numbers.
        """
        op_func = self._statistic_factory(np.amin)
        return op_func(self)

    def min_max_diff(self):
        r"""Calculate the difference between the min and max.

        .. math:: \max - \min

        The min-max diff is also called the range. It is
        a measure of spread.

        Returns:
          The difference between the min and the max.
        """
        return self.max() - self.min()

    def cov(self):
        r"""Calculate the pairwise covariances.

        The covariance is a measure of the relationship between two variables.
        The sign of the covariance indicates the direction of the relationship.

        .. math:: 

        \sigma_{XY} = \frac{1}{n} \sum_{i=1}^n (x_i - \mu_X) (y_i - \mu_Y)

        Returns:
          The pairwise covariances between all dimensions. This is usually
          a scalar when there are only 2 dimensions and a matrix when
          there are more than 2 dimensions.
        """
        op_func = self._multivariate_statistic_factory(
            lambda a: np.cov(a, rowvar=False, ddof=0)
        )
        return op_func(self)

    def corr(self):
        r"""Calculate the pairwise correlations.

        The correlation is the covariance normalized by the standard deviations.

        .. math:: 

        \rho_{XY} = \frac{1}{n} \sum_{i=1}^n \frac{x_i - \mu_X}{\sigma_X} \frac{y_i - \mu_Y}{\sigma_Y}

        Returns:
          The pairwise correlations between all dimensions. This is usually
          a scalar when there are only 2 dimensions and a matrix when
          there are more than 2 dimensions.
        """
        op_func = self._multivariate_statistic_factory(
            lambda a: np.corrcoef(a, rowvar=False, ddof=0)
        )
        return op_func(self)

    def corrcoef(self):
        r"""An alias for .corr()"""
        return self.corr()
    

class Logical:
    """A class that supports logical operations: and, or, and not.

    Subclasses must implement the _logical_factory method, which 
    specifies how the logical operator operates on two objects
    of that type.
    """

    def __and__(self, other):
        op_func = self._logical_factory(lambda x, y: x and y)
        return op_func(self, other)

    def __or__(self, other):
        op_func = self._logical_factory(lambda x, y: x or y)
        return op_func(self, other)

    def __invert__(self):
        op_func = self._logical_factory(lambda x: not x)
        return op_func(self)

    
class Filterable:
    """A class with filtering and counting methods.

    Subclasses must implement the filter method, which specifies how to 
    construct a new instance containing only those elements that satisfy 
    a given criterion.
    """

    def filter_eq(self, value):
        """Get all elements equal to a particular value.

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were equal to value.
        """
        return self.filter(lambda x: x == value)

    def filter_neq(self, value):
        """Get all elements _not_ equal to a particular value.

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were _not_ equal to value.
        """
        return self.filter(lambda x: x != value)

    def filter_lt(self, value):
        """Get all elements less than a particular value.

        N.B. lt stands for "less than". For elements that are 
        less than _or equal to_ the given value, use .filter_leq(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were less than value.
        """
        return self.filter(lambda x: x < value)

    def filter_leq(self, value):
        """Get all elements less than or equal to a particular value.

        N.B. leq stands for "less than or equal to". For elements 
        that are strictly less than the given value, use .filter_lt(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were less than _or equal to_ value.
        """
        return self.filter(lambda x: x <= value)

    def filter_gt(self, value):
        """Get all elements greater than a particular value.

        N.B. gt stands for "greater than". For elements that are 
        greater than _or equal to_ the given value, use .filter_geq(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were greater than value.
        """

        return self.filter(lambda x: x > value)

    def filter_geq(self, value):
        """Get all elements greater than or equal to a particular value.

        N.B. geq stands for "greater than or equal to". For elements
        that are strictly greater than the given value, use .filter_gt(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          All of the elements that were greater than _or equal to_ value.
        """
        return self.filter(lambda x: x >= value)


    # The following functions return an integer indicating
    # how many elements passed a given criterion.
    
    def count(self, func=lambda x: True):
        """Counts the number of elements satisfying a given criterion.

        Args:
          func (element -> bool): A function that takes in an element
            and returns a boolean (True/False). Only those elements
            that return True will be counted.

        Returns:
          int: The number of elements e for which func(e) is True.
        """
        return len(self.filter(func))

    def count_eq(self, value):
        """Count the number of elements equal to a particular value.

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were equal to value.
        """
        return len(self.filter_eq(value))

    def count_neq(self, value):
        """Count the number of elements _not_ equal to a particular value.

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were not equal to value.
        """
        return len(self.filter_neq(value))

    def count_lt(self, value):
        """Count the number of elements less than a particular value.

        N.B. lt stands for "greater than". For the number of elements
        that are less than _or equal to_ the given value, use 
        .count_leq(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were less than value.
        """
        return len(self.filter_lt(value))

    def count_leq(self, value):
        """Count the number of elements less than or equal to a particular value.

        N.B. leq stands for "less than or equal to". For the number of
        elements that are strictly greater than the given value, use 
        .count_lt(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were less than _or equal to_ value.
        """
        return len(self.filter_leq(value))

    def count_gt(self, value):
        """Count the number of elements greater than a particular value.

        N.B. gt stands for "greater than". For the number of elements
        that are greater than _or equal to_ the given value, use 
        .count_geq(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were greater than value.
        """
        return len(self.filter_gt(value))

    def count_geq(self, value):
        """Count the number of elements greater than or equal to a particular value.

        N.B. geq stands for "greater than or equal to". For the number of
        elements that are strictly greater than the given value, use 
        .count_gt(value).

        Args:
          value: A value of the same type as the elements in the object.
       
        Returns:
          int: The number of elements that were greater than _or equal to_ value.
        """
        return len(self.filter_geq(value))


class Transformable:
    """A class that supports transformations.

    Subclasses must implement the apply method, which specifies how to 
    apply a function to the object.
    """

    def __abs__(self):
        return self.apply(abs)

    def __round__(self):
        return self.apply(round)
    
    def __floor__(self):
        return self.apply(math.floor)

    def __ceil__(self):
        return self.apply(math.ceil)

    



"""Data structure for storing tabulated results.

This module defines a data structure, Table, that stores the
output of a .tabulate() operation.  Typically, Table stores
the possible outcomes and their counts or relative frequencies.
"""
from .base import Arithmetic


TABLE_TEMPLATE = '''
<table>
  <thead>
    <th width="80%">Outcome</th>
    <th width="20%">Value</th>
  </thead>
  <tbody>
    {table_body}
  </tbody>
</table>
'''


def _get_row_html(outcome, count):
    return "<tr><td>%s</td><td>%s</td></tr>" % (outcome, count)


class Table(dict, Arithmetic):

    def __init__(self, hash_map, outcomes=None):
        self.outcomes = outcomes
        if outcomes is None:
            for outcome, count in hash_map.items():
                self[outcome] = count
        else:
            for outcome in outcomes:
                self[outcome] = (
                    hash_map[outcome] if outcome in hash_map
                    else 0
                )

    def _repr_html_(self):
        # get keys in order
        if self.outcomes is None:
            keys = list(self.keys())
            try:
                keys.sort()
            except Exception:
                pass
        else:
            # preserve ordering of outcomes, if specified
            keys = self.outcomes

        # get HTML for table body
        table_body = ""
        for i, key in enumerate(keys):
            table_body += _get_row_html(key, self[key])
            # if we've already printed 19 rows, skip to end
            if i >= 18:
                table_body += _get_row_html("...", "...")
                table_body += _get_row_html(keys[-1], self[keys[-1]])
                break
        total = str(sum(self.values()))
        table_body += _get_row_html("<b>Total</b>", "<b>%s</b>" % total)

        # return HTML for entire table
        return TABLE_TEMPLATE.format(table_body=table_body)

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            return Table(
                {outcome: op(count, other) for outcome, count in self.items()},
                self.outcomes
            )

        return _op_func



from .distributions import Exponential
from .math import inf
from .probability_space import ProbabilitySpace
from .result import (
    InfiniteVector,
    ContinuousTimeFunction,
    DiscreteValued
)
from .random_variables import RV
from .random_processes import RandomProcess


class PoissonProcessResult(ContinuousTimeFunction,
                           DiscreteValued):

    def __init__(self, interarrival_times):
        self.interarrival_times = interarrival_times

        def func(t):
            total_time = 0
            for n, time in enumerate(self.interarrival_times):
                total_time += time
                if t < total_time:
                    return n

        super().__init__(func)

    def get_states(self):
        return InfiniteVector(lambda n: n)


class PoissonProcessProbabilitySpace(ProbabilitySpace):

    def __init__(self, rate):
        """Initialize probability space for a Poisson process.

        Args:
          rate: rate of the Poisson process
        """
        self.rate = rate

        def draw():
            interarrival_times = (Exponential(rate=self.rate) ** inf).draw()
            return PoissonProcessResult(interarrival_times)

        super().__init__(draw)


class PoissonProcess(RandomProcess, RV):

    def __init__(self, rate):
        """Initialize a Poisson process.

        Args:
          rate: rate of the Poisson process
        """
        self.rate = rate
        prob_space = PoissonProcessProbabilitySpace(self.rate)
        RandomProcess.__init__(self, prob_space)
        RV.__init__(self, prob_space)



"""Data structures for storing the results of a simulation.

This module provides data structures for storing the
results of a simulation, either outcomes from a
probability space or realizations of a random variable /
random process.
"""
import time

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.gridspec import GridSpec
from matplotlib.ticker import NullFormatter
from matplotlib.transforms import Affine2D

from .base import (Arithmetic, Statistical, Comparable,
                   Logical, Filterable, Transformable)
from .plot import (configure_axes, get_next_color, is_discrete,
                   count_var, compute_density, add_colorbar,
                   setup_ticks, make_tile, make_violin,
                   make_marginal_impulse, make_density2D)
from .result import (Scalar, Vector, TimeFunction,
                     is_number, is_numeric_vector)
from .table import Table


plt.style.use('seaborn-colorblind')


def _is_hashable(obj):
    return hasattr(obj, "__hash__")

def _is_boolean_vector(vector):
    return all(isinstance(x, (bool, np.bool_)) for x in vector)


class Results(Arithmetic, Statistical, Comparable,
              Logical, Filterable, Transformable):

    def __init__(self, results, sim_id=None):
        self.results = list(results)
        self.sim_id = time.time() if sim_id is None else sim_id

    def apply(self, func):
        """Apply a function to each outcome of a simulation.

        Args:
          func: A function to apply to each outcome.

        Returns:
          Results: A Results object of the same length,
            where each outcome is the result of applying
            the function to each outcome from the original
            Results object.
        """
        return type(self)(
            [func(result) for result in self.results],
            self.sim_id
        )

    def __getitem__(self, n):
        # if n is a Results object, use it as a boolean mask
        if isinstance(n, Results):
            return self.filter(n)
        # if n is a numeric array of values, return a Results
        # object with those dimensions
        elif is_numeric_vector(n):
            return self.apply(
                lambda result: type(result)(result[i] for i in n)
            )
        # otherwise, return the nth value of every simulation
        return self.apply(lambda result: result[n])

    def __iter__(self):
        for result in self.results:
            yield result

    def __len__(self):
        return len(self.results)

    def get(self, n):
        """Get the outcome of the nth simulation.

        Suppose x is an instance of a Results object.
        Although x behaves like a list, in that you
        can iterate over it, x[n] does not return
        the nth simulation. Instead, it returns a
        Results object, containing the nth dimension
        of every simulation. To get the outcome of the
        nth simulation, the .get(n) method is provided.

        Args:
          n (int): the index of the simulation result to get

        Returns:
          The outcome of the nth simulation.
        """

        # if n is a numeric array, return a Results object with those results
        if is_numeric_vector(n):
            return type(self)(
                self.results[i] for i in n
            )
        # otherwise, return the nth result (this also works when n is a slice)
        return self.results[n]

    def _get_counts(self):
        counts = {}
        for result in self.results:
            if _is_hashable(result):
                outcome = result
            elif isinstance(result, list) and all(_is_hashable(x) for x in result):
                outcome = tuple(result)
            else:
                outcome = str(result)
            if outcome in counts:
                counts[outcome] += 1
            else:
                counts[outcome] = 1
        return counts

    def tabulate(self, outcomes=None, normalize=False):
        """Counts up how much of each outcome there were.

        Args:
          outcomes (list): A list of outcomes to tabulate.
            By default, will tabulate all outcomes that
            appear in the Results.  Use this option if
            you want to include outcomes that could
            potentially not appear in the Results.
          normalize (bool): If True, return the relative
            frequency. Otherwise, return the counts.
            Defaults to False.

        Returns:
          Table: A Table with each of the observed
            outcomes and their freuencies.
        """
        table = Table(self._get_counts(), outcomes)
        if normalize:
            table /= len(self)
        return table

    # The Filterable superclass will use this to define all of the
    # .filter_*() and .count_*() methods.
    def filter(self, filt):
        """Get only the results that satisfy the given criterion.

        Args:
          filt: Either a function that takes in
            an outcome and returns a boolean, or
            a Results object of booleans of the
            same length as this Results object.

        Returns:
          Results: Another Results object containing
            only those outcomes corresponding to True.
        """
        if isinstance(filt, Results):
            if self.sim_id != filt.sim_id:
                raise Exception(
                    "In order to filter one Results object "
                    "by another, they must come from the "
                    "same simulation."
                )
            if len(filt) != len(self):
                raise ValueError(
                    "Filter must be the same length as the "
                    "Results object."
                )
            if not _is_boolean_vector(filt):
                raise ValueError(
                    "Every element in the filter must be a boolean."
                )
            return type(self)(x for x, cond in zip(self, filt) if cond)
        elif callable(filt):
            return type(self)(x for x in self if filt(x))
        else:
            raise TypeError(
                "A filter must be either a function or a "
                "boolean Results object of the same length."
            )

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            if isinstance(other, Results):
                if len(self) != len(other):
                    raise Exception(
                        "Results objects must be of the "
                        "same length."
                    )
                if self.sim_id != other.sim_id:
                    raise Exception(
                        "Results objects must come from the "
                        "same simulation."
                    )
                return type(self)(
                    [op(x, y) for x, y in zip(self, other)],
                    self.sim_id
                )
            else:
                return self.apply(lambda x: op(x, other))

        return _op_func

    # The Comparison superclass will use this to define all of the
    # usual comparison operations (e.g., <, >, ==, !=, etc.).
    def _comparison_factory(self, op):
        return self._operation_factory(op)

    # The Statistical superclass will use this to define all of the
    # usual comparison operations (e.g., <, >, ==, !=, etc.).
    def _statistic_factory(self, _):
        raise Exception("Statistical functions are only available "
                        "for simulations of random variables. "
                        "Define a RV on this probability space "
                        "and then try again.")

    def _multivariate_statistic_factory(self, op):
        self._statistic_factory(op)

    # The Logical superclass will use this to define the three
    # logical operations: and (&), or (|), not (~).
    def _logical_factory(self, op):

        def _op_func(self, other=None):
            # check that the vector only contains booleans
            if not _is_boolean_vector(self):
                raise ValueError(
                    "Logical operations are only defined for "
                    "boolean (True/False) Results objects.")
            # other will be None when op is the "not" operator
            if other is None:
                return Results([op(x) for x in self], self.sim_id)
            else:
                if isinstance(other, Results):
                    if self.sim_id != other.sim_id:
                        raise Exception("Results objects must come "
                                        "from the same simulation.")
                    if not _is_boolean_vector(other):
                        raise ValueError(
                            "Logical operations are only defined for "
                            "boolean (True/False) Results objects.")
                else:
                    raise TypeError(
                        "Logical operations are only defined "
                        "between two Results, not between a Result "
                        "and a %s." % type(other).__name__)
                return Results(
                    [op(x, y) for x, y in zip(self, other)],
                    self.sim_id
                )

        return _op_func


    def plot(self):
        raise Exception("Only simulations of random variables (RV) "
                        "can be plotted, but you simulated from a "
                        "probability space. You must first define a RV "
                        "on your probability space and simulate it. "
                        "Then call .plot() on those simulations.")

    def _repr_html_(self):

        table_template = '''
    <table>
      <thead>
        <th width="10%">Index</th>
        <th width="90%">Result</th>
      </thead>
      <tbody>
        {table_body}
      </tbody>
    </table>
        '''
        row_template = '''
        <tr>
          <td>%s</td><td>%s</td>
        </tr>
        '''

        def _truncate(result):
            if len(result) > 100:
                return result[:100] + "..."
            return result

        table_body = ""
        for i, result in enumerate(self.results):
            table_body += row_template % (i, _truncate(str(result)))
            # if we've already printed 9 rows, skip to end
            if i >= 8:
                table_body += "<tr><td>...</td><td>...</td></tr>"
                i_last = len(self) - 1
                table_body += row_template % (i_last, _truncate(str(self.get(i_last))))
                break
        return table_template.format(table_body=table_body)


class RVResults(Results):

    def __init__(self, results, sim_id=None):
        super().__init__(results, sim_id)
        # get type and dimension of the first result, if it exists
        iterresults = iter(self)
        try:
            first_result = next(iterresults)
        except StopIteration:
            return
        # determine the index set (if each realization is a TimeFunction)
        if isinstance(first_result, TimeFunction):
            self.index_set = first_result.index_set
        else:
            self.index_set = None
        # determine the dimension
        if is_number(first_result):
            self.dim = 1
        elif is_numeric_vector(first_result):
            self.dim = len(first_result)
        else:
            self.dim = None
        # iterate over remaining results, ensure they are consistent with the first
        for result in iterresults:
            if (isinstance(result, TimeFunction) and
                result.index_set != self.index_set):
                self.index_set = None
            if ((is_number(result) and self.dim != 1) or
                (is_numeric_vector(result) and self.dim != len(result))):
                self.dim = None

    def _set_array(self):
        # check if it has already been set
        if hasattr(self, "array"):
            return
        # don't set array for TimeFunctions
        elif self.index_set is not None:
            return
        # otherwise set array
        elif self.dim is not None:
            self.array = np.asarray(self.results)
        else:
            raise Exception(
                "This operation is only possible with results "
                "of consistent dimension.")

    # The Statistical superclass will use this to define all of the
    # usual comparison operations (e.g., <, >, ==, !=, etc.).
    def _statistic_factory(self, op):

        def _op_func(self):
            self._set_array()
            if self.dim == 1:
                return Scalar(op(a=self.array))
            elif self.dim is not None:
                return Vector(op(a=self.array, axis=0))
            elif self.index_set is not None:
                def _func(t):
                    return _op_func(self[t])
                return TimeFunction.from_index_set(self.index_set, _func)
            raise NotImplementedError(
                "Statistics can only be calculated for numerical "
                "data of consistent dimension."
            )

        return _op_func

    def _multivariate_statistic_factory(self, op):

        def _op_func(self):
            self._set_array()
            if self.dim == 2:
                return op(self.array)[0, 1]
            elif self.dim > 2:
                return op(self.array)
            elif self.dim == 1:
                raise Exception(
                    "This multivariate statistic is only defined when "
                    "when there are at least 2 dimensions.")
            raise NotImplementedError(
                "Statistics can only be calculated for numerical "
                "data of consistent dimension."
            )

        return _op_func

    def standardize(self):
        """Standardizes the results with respect to the mean and standard deviation.

        Returns:
          A new RVResults object, where every dimension has mean 0 and variance 1.
        """
        self._set_array()
        if self.dim is not None:
            return (self - self.mean()) / self.std()
        else:
            raise Exception("Could not standardize the given results.")

    def plot(self, type=None, alpha=None, normalize=True, jitter=False,
             bins=None, **kwargs):
        if type is not None:
            if isinstance(type, str):
                type = (type,)
            elif not isinstance(type, (tuple, list)):
                raise Exception("I don't know how to plot a " + str(type))

        if self.dim == 1:
            # make sure self.array, a Numpy array, has been set
            self._set_array()

            # determine plotting parameters
            counts = self._get_counts()
            discrete = is_discrete(counts.values())
            if type is None:
                type = ("impulse", ) if discrete else ("hist", )
            if alpha is None:
                alpha = .5
            if bins is None:
                bins = 30
            n = len(self)

            # initialize figure
            fig = plt.gcf()
            ax = plt.gca()
            color = get_next_color(ax)

            if 'density' in type:
                if discrete:
                    xs = sorted(list(counts.keys()))
                    probs = [counts[x] / n for x in xs]
                    ax.plot(xs, probs, marker='o', color=color, linestyle='-')
                    if len(type) == 1:
                        plt.ylabel('Relative Frequency')
                else:
                    density = compute_density(self.array)
                    xs = np.linspace(self.array.min(), self.array.max(), 1000)
                    ax.plot(xs, density(xs), linewidth=2, color=color)
                    if len(type) == 1 or (len(type) == 2 and 'rug' in type):
                        plt.ylabel('Density')

            if 'hist' in type or 'bar' in type:
                ax.hist(self.array, bins=bins, density=normalize,
                        color=color, alpha=alpha, **kwargs)
                plt.ylabel("Density" if normalize else "Count")
            elif 'impulse' in type:
                xs = list(counts.keys())
                freqs = list(counts.values())
                if normalize:
                    freqs = [freq / n for freq in freqs]
                if jitter:
                    a = .02 * (max(xs) - min(xs))
                    xs = [x + np.random.uniform(low=-a, high=a) for x in xs]
                # plot the impulses
                ax.vlines(xs, 0, freqs, color=color, alpha=alpha, **kwargs)
                configure_axes(ax, xs, freqs,
                               ylabel="Relative Frequency" if normalize else "Count")
            if 'rug' in type:
                xs = self.array
                if discrete:
                    noise_level = .002 * (self.array.max() - self.array.min())
                    xs = xs + np.random.normal(scale=noise_level, size=n)
                ax.plot(xs, [0.001] * n, '|', linewidth=5, color='k')
                if len(type) == 1:
                    setup_ticks([], [], ax.yaxis)
        elif self.dim == 2:
            # make sure self.array, a Numpy array, has been set
            self._set_array()
            x, y = self.array[:, 0], self.array[:, 1]

            x_count = count_var(x)
            y_count = count_var(y)
            x_height = x_count.values()
            y_height = y_count.values()
            discrete_x = is_discrete(x_height)
            discrete_y = is_discrete(y_height)

            if type is None:
                type = ("scatter",)
            if alpha is None:
                alpha = .5
            if bins is None:
                bins = 10 if 'tile' in type else 30

            if 'marginal' in type:
                fig = plt.gcf()
                gs = GridSpec(4, 4)
                ax = fig.add_subplot(gs[1:4, 0:3])
                ax_marg_x = fig.add_subplot(gs[0, 0:3])
                ax_marg_y = fig.add_subplot(gs[1:4, 3])
                color = get_next_color(ax)
                if 'density' in type:
                    densityX = compute_density(x)
                    densityY = compute_density(y)
                    x_lines = np.linspace(min(x), max(x), 1000)
                    y_lines = np.linspace(min(y), max(y), 1000)
                    ax_marg_x.plot(x_lines, densityX(x_lines), linewidth=2,
                                   color=get_next_color(ax))
                    ax_marg_y.plot(y_lines, densityY(y_lines), linewidth=2,
                                   color=get_next_color(ax),
                                   transform=Affine2D().rotate_deg(270) + ax_marg_y.transData)
                else:
                    if discrete_x:
                        make_marginal_impulse(x_count, get_next_color(ax), ax_marg_x, alpha, 'x')
                    else:
                        ax_marg_x.hist(x, color=get_next_color(ax), density=normalize,
                                       alpha=alpha, bins=bins)
                    if discrete_y:
                        make_marginal_impulse(y_count, get_next_color(ax), ax_marg_y, alpha, 'y')
                    else:
                        ax_marg_y.hist(y, color=get_next_color(ax), density=normalize,
                                       alpha=alpha, bins=bins, orientation='horizontal')
                plt.setp(ax_marg_x.get_xticklabels(), visible=False)
                plt.setp(ax_marg_y.get_yticklabels(), visible=False)
            else:
                fig = plt.gcf()
                ax = plt.gca()
                color = get_next_color(ax)

            nullfmt = NullFormatter() #removes labels on fig

            if 'scatter' in type:
                if jitter:
                    x = x + np.random.normal(loc=0, scale=.01 * (x.max() - x.min()), size=len(x))
                    y = y + np.random.normal(loc=0, scale=.01 * (y.max() - y.min()), size=len(y))
                ax.scatter(x, y, alpha=alpha, c=color, **kwargs)
            elif 'hist' in type:
                histo = ax.hist2d(x, y, bins=bins, cmap='Blues')

                # When normalize=True, use density instead of counts
                if normalize:
                    caxes = add_colorbar(fig, type, histo[3], 'Density')
                    #change scale to density instead of counts
                    new_labels = []
                    for label in caxes.get_yticklabels():
                        new_labels.append(int(label.get_text()) / len(x))
                    caxes.set_yticklabels(new_labels)
                else:
                    caxes = add_colorbar(fig, type, histo[3], 'Count')
            elif 'density' in type:
                den = make_density2D(x, y, ax)
                add_colorbar(fig, type, den, 'Density')
            elif 'tile' in type:
                hm = make_tile(x, y, bins, discrete_x, discrete_y, ax)
                add_colorbar(fig, type, hm, 'Relative Frequency')
            elif 'violin' in type:
                if discrete_x and not discrete_y:
                    positions = sorted(list(x_count.keys()))
                    make_violin(self.array, positions, ax, 'x', alpha)
                elif not discrete_x and discrete_y:
                    positions = sorted(list(y_count.keys()))
                    make_violin(self.array, positions, ax, 'y', alpha)
        else:
            if alpha is None:
                alpha = np.log(2) / np.log(len(self) + 1)
            ax = plt.gca()
            color = get_next_color(ax)
            for result in self.results:
                result.plot(alpha=alpha, color=color, **kwargs)
            plt.xlabel("Index")



import unittest
import numpy as np
import scipy.stats as stats

from symbulate import *

Nsim = 10000


class TestBernoulli(unittest.TestCase):

    def test_p_one(self):
        X = RV(Bernoulli(p=1))
        sims = X.sim(Nsim)
        self.assertTrue(all(sim == 1 for sim in sims))
   
    def test_sum(self):
        exp_list, obs_list = [], []
        X = RV(Bernoulli(p=.4) ** 5)
        sims = X.apply(sum).sim(Nsim)
        simulated = sims.tabulate()
        for k in range(6):
            expected = Nsim * stats.binom(n=5, p=.4).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > 0.01)

    def test_Bernoulli_Binomial_n_1(self):
        exp_list, obs_list = [], []   
        X = RV(Bernoulli(p=.4))
        sims = X.sim(Nsim)
        simulated = sims.tabulate()
        for k in range(2): 
            expected = Nsim * stats.binom(n=1, p=.4).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > 0.01)

        
class TestBinomial(unittest.TestCase):
    
    def test_Binomial_p_1(self):
        for nsample in range(1, 1000, 100):
            X = RV(Binomial(n=nsample, p=1.0))
            sims = X.sim(Nsim)
        self.assertTrue(all(sim == nsample for sim in sims))

    def test_Binomial_error_n(self):
        self.assertRaises(Exception, lambda: Binomial(n=-10, p=0.4))
    
    def test_Binomial_additive(self):
        exp_list, obs_list = [], []
        X, Y = RV(Binomial(n=8, p=0.6) * Binomial(n=5, p=0.6))
        sims = (X & Y).sim(Nsim).apply(sum)
        simulated = sims.tabulate()
        for k in range(2, 14):
            expected = Nsim * stats.binom(n=13, p=.6).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue   
        self.assertTrue(pval > 0.01)
    

class TestHypergeometric(unittest.TestCase):

    def test_Hypergeometric_no_failures(self):
        X = RV(Hypergeometric(n=10, N0=0, N1=1000))
        sims = X.sim(Nsim)
        self.assertTrue(all(sim == 10 for sim in sims))

    def test_Hypergeometric_Binomial_converge(self):
        exp_list, obs_list = [], []
        X = RV(Hypergeometric(n=8, N0=200, N1=800))
        sims = X.sim(Nsim)
        simulated = sims.tabulate()
        for k in range(9):
            expected = Nsim * stats.binom(n=8, p=.8).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > 0.01)

    def test_Hypergeometric_error_n_greater(self):
        self.assertRaises(Exception, 
                          lambda: Hypergeometric(n=10, N0=1, N1=8))


class TestGeometric(unittest.TestCase):
    
    def test_Geometric_error(self):
        self.assertRaises(Exception, lambda: Geometric(p=0))

    def test_Geometric_to_NBinom(self):
        exp_list, obs_list = [], []
        X = Geometric(p=0.8)
        sims = X.sim(Nsim)
        simulated = sims.tabulate()
        for k in range(1, 10):
            expected = Nsim  * stats.nbinom(n=1,p=0.8).pmf(k-1)
            if expected > 5:
                exp_list.append(expected) 
                obs_list.append(simulated[k]) 
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > 0.01)


class TestNegativeBinomial(unittest.TestCase):

    def test_NBinom_error_r(self):
        self.assertRaises(Exception, 
                          lambda: NegativeBinomial(r=-10, p=0.6))
    
    def test_NBinom_p_1(self):
        X = NegativeBinomial(r=10, p=1)
        sims = X.sim(Nsim)
        self.assertTrue(all(sim == 10 for sim in sims))
    
    def test_NBinom_Pascal_additive(self):
        exp_list, obs_list = [], []
        X, Y = RV(Pascal(r=4, p=0.6) * Pascal(r=6, p=0.6))
        sims = (X + Y).sim(Nsim)
        simulated = sims.tabulate()
        for k in range(10, 35):
            expected = Nsim * stats.nbinom(n=10, p=0.6).pmf(k)
            if expected > 5:
                exp_list.append(expected) 
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > .01)

    def test_NBinom_to_Geometric(self):
        exp_list, obs_list = [], []
        X = NegativeBinomial(r=1, p=0.8)
        sims = X.sim(Nsim)
        simulated = sims.tabulate()
        for k in range(10):
            expected = Nsim * stats.geom(p=0.8).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval  = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > 0.01)


class TestPascal(unittest.TestCase):

    def test_Pascal_error_r(self):
        self.assertRaises(Exception, lambda: Pascal(r=0, p=0.3))
    
    def test_Pascal_p_1(self):
        X = Pascal(r=10, p=1.0)
        sims = X.sim(Nsim)
        self.assertTrue(all(sim == 0 for sim in sims))


class TestPoisson(unittest.TestCase):
    
    def test_Poisson_error(self):
        self.assertRaises(Exception, lambda: Poisson(lam=0))

    def test_Poisson_additive(self):
        exp_list, obs_list = [], []
        X, Y = RV(Poisson(lam=4) * Poisson(lam=7))
        sims = (X + Y).sim(Nsim)
        simulated = sims.tabulate()
        for k in range(25):
            expected = Nsim * stats.poisson(mu=11).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > .01)

    def test_conditional_Poisson_add(self):
        obs_list, exp_list = [], []
        X,Y = RV(Poisson(lam=6) * Poisson(lam=7))
        sims = (X|(X+Y == 12)).sim(Nsim)
        simulated = sims.tabulate()
        for k in range(12):
            expected = Nsim * stats.binom(n=12, p=6 / 13).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > .01)


class TestUniform(unittest.TestCase):
    
    def test_Uniform_error(self):
        self.assertRaises(Exception, lambda: Uniform(a=6, b=-1))

    def test_conditional_exp_uniform(self):
        X, Y = RV(Exponential(rate=3) ** 2)
        sims = (X | (X < 3) & (X + Y > 3)).sim(1000)
        cdf = stats.uniform(loc=0, scale=3).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Uniform_to_ChiSquare(self):
        X = RV(Uniform(a=0, b=1))
        sims = (-2 * log(X)).sim(Nsim)
        cdf = stats.chi2(df=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Uniform_to_Exponential(self):
        X = RV(Uniform(a=0, b=1))
        Y = -1 / 5 * log(X)
        sims = Y.sim(Nsim)
        cdf = stats.expon(scale=1/5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Uniform_to_Beta(self):
        X = RV(Uniform(a=0, b=1))
        sims = (X ** 15).sim(Nsim)
        cdf = stats.beta(a=1/15, b=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Uniform_to_Cauchy(self):
        X = RV(Uniform(a=0, b=1))
        sims = (pi * (X - 1/2)).apply(tan).sim(Nsim)
        cdf = stats.cauchy(loc=0, scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Uniform_to_Pareto(self):
        X = RV(Uniform(a=0, b=1))
        sims = (2 * X ** (-1 / 0.1)).sim(10000)
        cdf = stats.pareto(b=0.1, loc=0, scale=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)
        
class TestNormal(unittest.TestCase):

    def test_Normal_error(self):
        self.assertRaises(Exception, lambda: Normal(mean=0, var=-10)) 
    
    def test_sum(self):
        X = RV(Normal(mean=-1, sd=2) ** 3)
        sims = X.apply(sum).sim(Nsim)
        cdf = stats.norm(loc=-3,
                         scale=np.sqrt(12)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_sum_Standard_Normal(self):
        X, Y = RV(Normal(mean=0, var=1) ** 2)
        sims = (X + Y).sim(Nsim)
        cdf = stats.norm(loc=0, scale=sqrt(2)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_subtract_Standard_Normal(self):
        X, Y = RV(Normal(mean=0, var=1) ** 2)
        sims = (X - Y).sim(Nsim)
        cdf = stats.norm(loc=0, scale=sqrt(2)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_standardize(self):
        X = RV(Normal(mean=8, var=4))
        X_stand = (X - 8) / 2
        sims = X_stand.sim(Nsim)
        cdf = stats.norm(loc=0, scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_standardize_to_Normal(self):
        Z = RV(Normal(mean=0, sd=1))
        X = 10 + 5 * Z
        sims = X.sim(Nsim)
        cdf = stats.norm(loc=10, scale=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_to_Gamma(self):
        X = RV(Normal(mean=0, var=1))
        X = X ** 2
        sims = X.sim(Nsim)
        cdf = stats.gamma(a=1/2, scale=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_to_ChiSquare(self):
        X = RV(Normal(mean=0, var=1))
        X = X ** 2
        sims = X.sim(Nsim)
        cdf = stats.chi2(df=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_to_Cauchy(self):
        X, Y = RV(Normal(mean=0, var=1) ** 2)
        sims = (X / Y).sim(Nsim)
        cdf = stats.cauchy(loc=0, scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_to_F(self):
        A, B, C, V, W, X, Y, Z = RV(Normal(mean=0, var=1) ** 8)
        sims = ((((A ** 2) + (B ** 2) + (C ** 2)) / 3) /
               (((V ** 2) + (W ** 2) + (X ** 2) + (Y ** 2) + (Z ** 2)) 
                                                            / 5)).sim(Nsim)
        cdf = stats.f(dfn=3, dfd=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_sum_Normal_to_ChiSquare(self):
        X, Y, Z, A, B = RV(Normal(mean=0, var=1) ** 5)
        sims = ((X ** 2) + (Y ** 2) + (Z ** 2) + (A ** 2) + (B ** 2)).sim(Nsim)
        cdf = stats.chi2(df=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

 
class TestExponential(unittest.TestCase):
    
    def test_Exponential_error(self):
        self.assertRaises(Exception, lambda: Exponential(rate=-5))

    def test_Exponential_to_Gamma(self):
        X = RV(Exponential(rate=0.9))
        sims = X.sim(Nsim)
        cdf = stats.gamma(scale=1/0.9, a=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Exponential_sum_Gamma(self):
        X, Y, Z, A = RV(Exponential(rate=0.9) ** 4)
        sims = (X + Y + Z + A).sim(Nsim)
        cdf = stats.gamma(scale=1/0.9, a=4).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Exponential_to_ChiSquare(self):
        X = RV(Exponential(rate=1/2))
        sims = X.sim(Nsim)
        cdf = stats.chi2(df=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Exponential_to_Pareto(self):
        X = RV(Exponential(rate=2))
        sims = (3 * exp(X)).sim(Nsim)
        cdf = stats.pareto(b=2, loc=0, scale=3).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Exponential_to_Weibull(self):
        X = RV(Exponential(rate=5))
        sims = X.sim(Nsim)
        cdf= stats.frechet_r(scale=1/5, c=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Exponential_to_Rayleigh(self):
        X = RV(Exponential(rate=5))
        sims = sqrt(X).sim(Nsim)
        cdf = stats.rayleigh(scale=1 / sqrt(2 * 5)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Poisson_Exponential_to_Geometric(self):
        def poisson_exp():
            x = Exponential(rate=1 / lam).draw() 
            z = RV(Poisson(x)).draw()
            return z

        exp_list, obs_list, lam = [], [], 1/5
        P = ProbabilitySpace(poisson_exp)
        A = RV(P)
        sims = (A + 1).sim(Nsim)
        simulated = sims.tabulate()
        for k in range(40):
            expected = Nsim * stats.geom(p=1 / (1 + lam)).pmf(k)
            if expected > 5:
                exp_list.append(expected)
                obs_list.append(simulated[k])
        pval = stats.chisquare(obs_list, exp_list).pvalue
        self.assertTrue(pval > .01)


class TestGamma(unittest.TestCase):

    def test_Gamma_shape_error(self):
        self.assertRaises(Exception, lambda: Gamma(shape=-5, rate=40))

    def test_Gamma_rate_error(self):
        self.assertRaises(Exception, lambda: Gamma(shape=4, rate=-10))

    def test_Gamma_to_Exponential(self):
        X = Gamma(shape=1, rate=1/ 0.9)
        sims = X.sim(Nsim)
        cdf = stats.expon(scale=0.9).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)
        
    def test_Gamma_reshape(self):
        X = RV(Gamma(shape=9, scale=4))
        sims = (X * 8).sim(Nsim)
        cdf = stats.gamma(scale=4 * 8, a=9).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)
 
    def test_Gamma_additive(self):
        X, Y = RV(Gamma(shape=10, scale=0.5) * 
                  Gamma(shape=8, scale=0.5))
        sims = (X + Y).sim(Nsim)
        cdf = stats.gamma(scale=0.5, a=18).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Gamma_to_Beta(self):
        X, Y = RV(Gamma(shape=5, scale=8) * Gamma(shape=4, scale=8))
        sims = (X / (X + Y)).sim(Nsim)
        cdf = stats.beta(a=5, b=4).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Gamma_to_F(self):
        X, Y = RV(Gamma(shape=2, rate=5) * Gamma(shape=4, rate=7))
        sims = ((4 * 5 * X) / (2 * 7 * Y)).sim(Nsim)
        cdf = stats.f(dfn=2 * 2, dfd=2 * 4).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Gamma_to_ChiSquare(self):
        X = RV(Gamma(shape=10/2, scale=2))
        sims = X.sim(Nsim)
        cdf = stats.chi2(df=10).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)


class TestBeta(unittest.TestCase):
    
    def test_Beta_error_a(self):
        self.assertRaises(Exception, lambda: Beta(a=-10, b=3))

    def test_Beta_error_b(self):
        self.assertRaises(Exception, lambda: Beta(a=3, b=-10))

    def test_Beta_to_Uniform(self):
        X = Beta(a=1, b=1)
        sims = X.sim(Nsim)
        cdf = stats.uniform(loc=0, scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Beta_symmetry(self):
        X = RV(Beta(a=4, b=5))
        sims = (1 - X).sim(Nsim)
        cdf = stats.beta(a=5, b=4).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Beta_to_Exponential(self):
        X = RV(Beta(a=0.7, b=1))
        sims = (-log(X)).sim(Nsim)
        cdf = stats.expon(scale=1 / 0.7).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Beta_to_F(self):
        X = RV(Beta(a=10/2, b=12/2))
        sims = (12 * X / (10 * (1 - X))).sim(Nsim)
        cdf = stats.f(dfn=10, dfd=12).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)


class TestStudentT(unittest.TestCase):
    
    def test_StudentT_df_error(self):
        self.assertRaises(Exception, lambda: StudentT(df=0))

    def test_StudentT_to_Normal(self):
        X = StudentT(df=Nsim)
        sims = X.sim(Nsim)
        cdf = stats.norm(loc=0,scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_ChiSquare_to_StudentT(self):
        X, Y = RV(Normal(mean=0, var=1) * ChiSquare(df=5))
        sims = (X / sqrt(Y / 5)).sim(Nsim)
        cdf = stats.t(df=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)    


class TestChiSquare(unittest.TestCase):
    
    def test_ChiSquare_error(self):
        self.assertRaises(Exception, lambda: ChiSquare(df=0.5))

    def test_ChiSquare_to_Gamma(self):
        X = RV(ChiSquare(df=10))
        sims = X.sim(Nsim)
        cdf = stats.gamma(a=5, scale=1/0.5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_ChiSquare_to_F(self):
        X, Y = RV(ChiSquare(df=3) * ChiSquare(df=5))
        sims = ((X / 3) / (Y / 5)).sim(Nsim)
        cdf = stats.f(dfn=3, dfd=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_ChiSquare_to_Beta(self):
        X, Y = RV(ChiSquare(df=4) * ChiSquare(df=5))
        sims = (X / (X + Y)).sim(Nsim)
        cdf = stats.beta(a=4/2, b=5/2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)
    

class TestF(unittest.TestCase):

    def test_F_error(self):
        self.assertRaises(Exception, lambda: F(dfN=0, dfD=5))

    def test_inverse_T(self):
        X = RV(F(dfN=4, dfD=8))
        sims = (1 / X).sim(Nsim)
        cdf = stats.f(dfn=8, dfd=4).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_StudentT_to_F(self):
        X = RV(StudentT(df=15))
        sims = (X ** 2).sim(Nsim)
        cdf = stats.f(dfn=1, dfd=15).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_F_to_Beta(self):
        X = RV(F(dfN=5, dfD=8))
        sims = ((5 * X / 8) / (1 + (5 * X / 8))).sim(Nsim)
        cdf = stats.beta(a=5/2, b=8/2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)


class TestCauchy(unittest.TestCase):

    def test_Cauchy_mean(self):
        X = Cauchy()
        math.isnan(X.mean())

    def test_Cauchy_to_T(self):
        X = RV(Cauchy())
        sims = X.sim(Nsim)
        cdf = stats.t(df=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Cauchy_inverse(self):
        X = RV(Cauchy())
        sims = (1 / X).sim(Nsim)
        cdf = stats.cauchy(loc=0, scale=1).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)

    def test_Cauchy_additive(self):
        X, Y = RV(Cauchy()**2)
        sims = (X + Y).sim(Nsim)
        cdf = stats.cauchy(loc=0, scale=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > 0.01)


class TestLognormal(unittest.TestCase):

    def test_LogNormal_error(self):
        self.assertRaises(Exception, lambda: LogNormal(mu=0, sigma=-5)) 

    def test_LogNormal_to_Normal(self):
        X = LogNormal(mu=10, sigma=5)
        sims = X.sim(Nsim).apply(log)
        cdf = stats.norm(loc=10, scale=5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Normal_to_LogNormal(self):
        X = RV(Normal(mean=10, sd=5))
        sims = X.apply(exp).sim(Nsim)
        cdf = stats.lognorm(s=5, scale=exp(10)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_LogNormal_Product(self):
        X, Y = RV(LogNormal(mu=10, sigma=5) * LogNormal(mu=11, sigma=6))
        sims = (X * Y).sim(Nsim)
        cdf = stats.lognorm(s=sqrt(25 + 36), scale=exp(21)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

class TestPareto(unittest.TestCase):

    def test_Pareto_check_mean(self):
        x = stats.pareto(b=-3) 
        math.isnan(x.mean())

    def test_Pareto_to_Exponential(self):
        X = RV(Pareto(b=1.5, scale=0.1))
        sims = (log(X / 0.1)).sim(Nsim)
        cdf = stats.expon(scale=1/1.5).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)


class TestWeibull(unittest.TestCase):
    def test_Weibull_to_Exponential(self):
        '''X = RV(Weibull(scale=1, c=10))
           sims = X.sim(Nsim)
           cdf = stats.expon(scale=1/10).cdf
           pval = stats.kstest(sims, cdf).pvalue
           self.assertTrue(pval > .01)
        '''
        pass

class TestRayleigh(unittest.TestCase):

    def test_Rayleigh_Normal(self):
        A, B = RV(Normal(mean=0, var=1) * Normal(mean=0, var=1))
        sims = (A ** 2 + B ** 2).apply(sqrt).sim(Nsim)
        cdf = stats.rayleigh.cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)

    def test_Rayleigh_to_Chi(self):
        X = RV(Rayleigh())
        sims = X.sim(Nsim)
        cdf = stats.chi(df=2).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)


class TestMultivariateNormal(unittest.TestCase):
     
    def test_MultivariateNormal_mean_cov_error(self):
        self.assertRaises(Exception, 
             lambda: MultivariateNormal(mean=[2], cov=[[2,2], [3,4]]))

    def test_MultivariateNormal_cov_square_error(self):
        self.assertRaises(Exception, 
             lambda: MultivariateNormal(mean=[2,4], cov=[[2,4,5],[2,1]]))


class TestBivariateNormal(unittest.TestCase):
    
    def test_BivariateNormal_error1(self):
        self.assertRaises(Exception, lambda: 
             BivariateNormal(mean1=3, mean2=4, sd1=-3, sd2=3, corr=0.9))

    def test_BivariateNormal_error2(self):
        self.assertRaises(Exception, lambda: 
             BivariateNormal(mean1=3, mean2=4, sd1=3, sd2=3, corr=1.1))
 
    def test_LinCom_BivNormal(self):
        X, Y = RV(BivariateNormal(mean1=30, mean2=50, sd1=8, sd2=6, corr=-0.4))
        Z = 4 * X - 2 * Y
        Z_mean = 4 * 30 + (-2) * 50
        Z_var = 16 * 64 + 4 * 36 + 2 * (4) * (-2) * (-0.4 * 8 * 6)
        sims = Z.sim(Nsim)
        cdf = stats.norm(loc=Z_mean, scale=sqrt(Z_var)).cdf
        pval = stats.kstest(sims, cdf).pvalue
        self.assertTrue(pval > .01)
    
    def test_BivNormal_condDistr_r(self):
        for c in [-0.9, 0.9, 0.1]:
            X,Y = RV(BivariateNormal(mean1=20, mean2=10, sd1=3, sd2=5, corr=c))
            sims = (Y | (abs(X - 21) < 0.1)).sim(500)
            cdf = stats.norm(loc=10 + c * 5 / 3 * (21 - 20), 
                             scale=5 * sqrt(1 - c ** 2)).cdf
            pval = stats.kstest(sims, cdf).pvalue
            self.assertTrue(pval > 0.01)





from .probability_space import ProbabilitySpace, BoxModel, DeckOfCards
from .random_variables import RV
from .random_processes import RandomProcess
from .distributions import (
    Bernoulli,
    Binomial,
    Hypergeometric,
    Geometric,
    NegativeBinomial,
    Pascal,
    Poisson,
    DiscreteUniform,
    Uniform,
    Normal,
    Exponential,
    Gamma,
    Beta,
    StudentT,
    ChiSquare,
    F,
    Cauchy,
    LogNormal,
    Pareto,
    Rayleigh,
    MultivariateNormal,
    BivariateNormal,
    Multinomial
)
from .independence import AssumeIndependent
from .index_sets import (
    Naturals,
    Integers,
    Reals,
    DiscreteTimeSequence
)
from .result import (
    Scalar,
    Vector,
    InfiniteVector,
    DiscreteTimeFunction,
    ContinuousTimeFunction,
    concat
)
from .gaussian_process import (
    GaussianProcess,
    GaussianProcessProbabilitySpace,
    BrownianMotion,
    BrownianMotionProbabilitySpace
)
from .poisson_process import (
    PoissonProcess,
    PoissonProcessProbabilitySpace
)
from .markov_chains import (
    MarkovChain,
    MarkovChainProbabilitySpace,
    ContinuousTimeMarkovChain,
    ContinuousTimeMarkovChainProbabilitySpace
)
from .plot import figure, xlabel, ylabel, xlim, ylim, plot
from .math import *



import numbers
import numpy as np
import matplotlib.pyplot as plt

import symbulate
from .base import Arithmetic, Transformable, Statistical, Filterable
from .index_sets import DiscreteTimeSequence, Reals, Naturals


class Scalar(numbers.Number):

    def __new__(cls, value, *args, **kwargs):
        if isinstance(value, numbers.Integral):
            return Int(value)
        elif isinstance(value, (float, np.floating)):
            return Float(value)
        else:
            raise Exception("Scalar type not understood.")


class Int(int, Scalar):

    def __new__(cls, value, *args, **kwargs):
        return super(Int, cls).__new__(cls, value)


class Float(float, Scalar):

    def __new__(cls, value, *args, **kwargs):
        return super(Float, cls).__new__(cls, value)


class Tuple(Arithmetic, Transformable, Statistical, Filterable):
    """A collapsible data structure.
    """

    def __init__(self, values):
        if is_scalar(values):
            self.values = (values, )
        elif hasattr(values, "__len__") or hasattr(values, "__next__"):
            self.values = tuple(values)
        else:
            raise Exception(
                "Tuples can only be created from "
                "finite iterable data."
            )

    def __getitem__(self, n):
        # if n is a numeric array, return a Tuple of those values
        if is_numeric_vector(n):
            return type(self)(self.values[i] for i in n)
        # otherwise, return the value at n
        return self.values[n]

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        for value in self.values:
            yield value

    def __hash__(self):
        return hash(tuple(self.values))

    # Define comparison operators to handle sorting.
    def __eq__(self, other):
        if not hasattr(other, "__len__"):
            return False
        if len(self) != len(other):
            return False
        return all(a == b for a, b in zip(self, other))

    def __lt__(self, other):
        return tuple(self.values) < tuple(other.values)

    def apply(self, func):
        """Apply function to every element of a Tuple.

        Args:
          func: function to apply to the Tuple

        Example:
          x = Tuple([1, 2, 3])
          y = x.apply(log)

        Note: For most standard functions, you can apply the function to
          the Tuple directly. For example, in the example above,
          y = log(x) would have been equivalent and more readable.

        User defined functions can also be applied.

        Example:
          def log_squared(n):
            return log(n) ** 2
          y = x.apply(log_squared)
        """
        return type(self)(func(e) for e in self)

    # The Filterable superclass will use this to define all of the
    # .filter_*() and .count_*() methods.
    def filter(self, filt):
        """Get only the elements that satisfy the given criterion.

        Args:
          filt: A function that takes in an element and returns
            a boolean.

        Returns:
          Tuple: Another Tuple containing only those elements e
          where filt(e) is True.
        """
        return type(self)(e for e in self if filt(e))

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            if is_number(other):
                return type(self)(op(value, other) for value in self)
            elif is_vector(other):
                # check that other is the same length as the Tuple
                if len(self) != len(other):
                    raise Exception(
                        "Arithmetic operations between a %s and a %s "
                        "are only valid if they are the same length. "
                        "You attempted to combine a %s of length %d "
                        "with a %s of length %d." % (
                            type(self).__name__,
                            type(other).__name__,
                            type(self).__name__,
                            len(self),
                            type(other).__name__,
                            len(other)
                        ))
                # return a new Tuple/Vector of the same length
                return type(self)(op(a, b) for a, b in zip(self, other))
            else:
                return NotImplemented

        return _op_func

    # The Statistical superclass will use this to define all of the
    # usual statistical functions (e.g., mean, var, etc.)
    def _statistic_factory(self, op):
        def _op_func(self):
            return op(self.values)
        return _op_func

    def cumsum(self):
        return type(self)(np.cumsum(self.values))

    def plot(self, **kwargs):
        plt.plot(range(len(self)), self.values, '.--', **kwargs)

    def __str__(self):
        if len(self) <= 6:
            return "(" + ", ".join(str(x) for x in self) + ")"
        else:
            first_few = ", ".join(str(x) for x in self[:5])
            last = str(self[-1])
            return "(" + first_few + ", ..., " + last + ")"

    def __repr__(self):
        return self.__str__()


class Vector(Tuple):
    """A data structure like a Tuple, except it does not collapse.
    """
    pass


class TimeFunction(Arithmetic):

    @classmethod
    def from_index_set(cls, index_set, func=None):
        if isinstance(index_set, DiscreteTimeSequence):
            return DiscreteTimeFunction(func, index_set=index_set)
        elif isinstance(index_set, Reals):
            return ContinuousTimeFunction(func)
        elif isinstance(index_set, Naturals):
            return InfiniteVector(func)

    def check_same_index_set(self, other):
        if isinstance(other, (numbers.Number, symbulate.RV)):
            return
        elif isinstance(other, TimeFunction):
            if self.index_set != other.index_set:
                raise Exception(
                    "Operations can only be performed on "
                    "TimeFunctions with the same index set."
                )
        else:
            raise Exception("Cannot combine %s with %s." % (
                type(self).__name__, type(other).__name__
            ))


class InfiniteTuple(TimeFunction):

    def __init__(self, func=lambda n: n):
        """Initializes a (lazy) data structure for an infinite vector.

        Args:
          func: A function of n that returns the value in position n.
                n is assumed to be a natural number (integer >= 0).
                This function can be defined at initialization time,
                or later. By default, it is not set at initialization.
        """
        if func is not None:
            self.func = func
        self.index_set = Naturals()
        self.values = []

    def __getitem__(self, n):
        m = len(self.values)
        # Add necessary elements to self.values
        n0 = None
        # handle the case where n is a slice
        if isinstance(n, slice):
            if n.stop is None:
                if n.start is None:
                    return self
                else:
                    return type(self)(lambda i: self[i + n.start])
            if n.stop >= m:
                n0 = n.stop
        elif isinstance(n, numbers.Integral) and n >= m:
            n0 = n
        if n0 is not None:
            for i in range(m, n0 + 1):
                self.values.append(self.func(i))
        # Return the corresponding value(s)
        return self.values[n]

    def __call__(self, n):
        return self[n]

    def __str__(self):
        first_few = [str(self[i]) for i in range(6)]
        return "(" + ", ".join(first_few) + ", ...)"

    def __repr__(self):
        return self.__str__()

    def apply(self, func):
        """Apply function to every element of an InfiniteTuple.

        Args:
          func: function to apply to the InfiniteTuple

        Example:
          x = InfiniteTuple(lambda n: n)
          y = x.apply(log)

        Note: For most standard functions, you can apply the function to
          the InfiniteTuple directly. For example, in the example above,
          y = log(x) would have been equivalent and more readable.

        User defined functions can also be applied.

        Example:
          def log_squared(n):
            return log(n) ** 2
          y = x.apply(log_squared)
        """
        return type(self)(lambda n: func(self[n]))

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            self.check_same_index_set(other)
            if is_number(other):
                return type(self)(lambda n: op(self[n], other))
            elif isinstance(other, InfiniteTuple):
                return type(self)(lambda n: op(self[n], other[n]))
            else:
                return NotImplemented

        return _op_func


class InfiniteVector(InfiniteTuple):

    def cumsum(self):
        def _func(n):
            return sum(self[i] for i in range(n + 1))
        return InfiniteVector(_func)

    def plot(self, tmin=0, tmax=10, **kwargs):
        xs = range(tmin, tmax)
        ys = [self[t] for t in range(tmin, tmax)]
        plt.plot(xs, ys, '.--', **kwargs)


class DiscreteTimeFunction(TimeFunction):

    def __init__(self, func=None, fs=1, index_set=None):
        """Initializes a data structure for a discrete-time function.

        Args:
          func: A function of n that returns the value at time n / fs.
            n is assumed to be any integer (postive or negative).
            By default, it is set to the identity function f[n] = n / fs.
          fs (int): The sampling rate of the function, in Hertz (samples
            per second).
          index_set (IndexSet): The index set of the discrete-time function
            (fs is ignored if this is specified.)
        """
        if func is not None:
            self.func = func
        else:
            self.func = lambda n: n / fs
        if index_set is None:
            self.index_set = DiscreteTimeSequence(fs)
        else:
            self.index_set = index_set
        self.array_pos = [] # stores values for t >= 0
        self.array_neg = [] # stores values for t < 0

    def _get_value_at_index(self, n):
        if not isinstance(n, numbers.Integral):
            raise KeyError(
                "For a DiscreteTimeFunction f, f[n] returns the "
                "the nth time sample, so n must be an integer. "
                "If you want the value at time t, try f(t) instead.")

        if n >= 0:
            m = len(self.array_pos)
            if n >= m:
                for i in range(m, n + 1):
                    self.array_pos.append(self.func(i))
            return self.array_pos[n]
        else:
            m = len(self.array_neg)
            if -n > m:
                for i in range(-m - 1, n - 1, -1):
                    self.array_neg.append(self.func(i))
            return self.array_neg[-n - 1]

    def _get_value_at_time(self, t):
        fs = self.index_set.fs
        if not t in self.index_set:
            raise KeyError((
                "No value at time %.2f for a function with "
                "a sampling rate of %d Hz.") % (t, fs))
        return self._get_value_at_index(int(t * fs))

    def __getitem__(self, n):
        if is_number(n):
            return self._get_value_at_index(n)
        elif is_numeric_vector(n):
            return Vector(self._get_value_at_index(e) for e in n)
        elif isinstance(n, slice):
            return Vector(self._get_value_at_index(e) for e in
                          range(n.start, n.stop, n.step or 1))
        else:
            raise TypeError("Cannot evaluate DiscreteTimeFunction at "
                            "index %s (type %s)." % (n, type(n).__name__))

    def __call__(self, t):
        if is_number(t):
            return self._get_value_at_time(t)
        elif is_numeric_vector(t):
            return Vector(self._get_value_at_time(e) for e in t)
        elif isinstance(t, DiscreteTimeFunction):
            self.check_same_index_set(t)
            return DiscreteTimeFunction(func=lambda n: self(t[n]),
                                        index_set=self.index_set)
        else:
            raise TypeError("Cannot evaluate DiscreteTimeFunction at "
                            "time %s (type %s)." % (t, type(t).__name__))

    def apply(self, func):
        """Compose function with the TimeFunction.

        Args:
          func: function to compose with the TimeFunction

        Example:
          f = DiscreteTimeFunction(lambda t: t, fs=1)
          g = f.apply(log)

        Note: For most standard functions, you can apply the function to
          the TimeFunction directly. For example, in the example above,
          g = log(f) would have been equivalent and more readable.

        User-defined functions can also be applied.

        Example:
          def log_squared(f):
            return log(f) ** 2
          g = f.apply(log_squared)
        """
        return DiscreteTimeFunction(lambda n: func(self[n]),
                                    index_set=self.index_set)

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            self.check_same_index_set(other)
            if is_number(other):
                return DiscreteTimeFunction(
                    lambda n: op(self[n], other),
                    index_set=self.index_set
                )
            elif isinstance(other, DiscreteTimeFunction):
                return DiscreteTimeFunction(
                    lambda n: op(self[n], other[n]),
                    index_set=self.index_set
                )
            else:
                return NotImplemented

        return _op_func

    def __str__(self):
        first_few = ", ".join(str(self[n]) for n in range(-2, 3))
        return "(..., " + first_few + ", ...)"

    def __repr__(self):
        return self.__str__()

    def plot(self, tmin=0, tmax=10, **kwargs):
        nmin = int(np.floor(tmin * self.index_set.fs))
        nmax = int(np.ceil(tmax * self.index_set.fs))
        ts = [self.index_set[n] for n in range(nmin, nmax)]
        ys = [self[n] for n in range(nmin, nmax)]
        plt.plot(ts, ys, ".--", **kwargs)


class ContinuousTimeFunction(TimeFunction):

    def __init__(self, func=lambda t: t):
        """Initializes a data structure for a discrete-time function.

        Args:
          func: A function of n that returns the value in position n.
                n is assumed to be any integer (postive or negative).
                This function can be defined at initialization time,
                or later. By default, it is not set at initialization.
        """
        self.index_set = Reals()
        if func is not None:
            self.func = func

    def __call__(self, t):
        if is_number(t):
            return self.func(t)
        elif is_numeric_vector(t):
            try:
                # Use vectorized function if it exists
                return Vector(self.vfunc(t))
            except:
                return Vector(self.func(e) for e in t)
        elif isinstance(t, ContinuousTimeFunction):
            return ContinuousTimeFunction(func=lambda s: self(t(s)))
        else:
            raise TypeError("Cannot evaluate ContinuousTimeFunction at "
                            "time %s (type %s)." % (t, type(t).__name__))

    def __getitem__(self, t):
        return self(t)

    def apply(self, func):
        """Compose function with the TimeFunction.

        Args:
          func: function to compose with the TimeFunction


        Example:
          f = ContinuousTimeFunction(lambda t: t)
          g = f.apply(log)

        Note: For most standard functions, you can apply the function to
          the TimeFunction directly. For example, in the example above,
          g = log(f) would have been equivalent and more readable.

        User-defined functions can also be applied.

        Example:
          def log_squared(f):
            return log(f) ** 2
          g = f.apply(log_squared)
        """
        return ContinuousTimeFunction(lambda t: func(self(t)))

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.).
    def _operation_factory(self, op):

        def _op_func(self, other):
            self.check_same_index_set(other)
            if is_number(other):
                return ContinuousTimeFunction(
                    lambda t: op(self(t), other)
                )
            elif isinstance(other, ContinuousTimeFunction):
                return ContinuousTimeFunction(
                    lambda t: op(self(t), other(t))
                )
            else:
                return NotImplemented

        return _op_func

    def __str__(self):
        return "[continuous-time function]"

    def __repr__(self):
        return self.__str__()

    def plot(self, tmin=0, tmax=10, **kwargs):
        ts = np.linspace(tmin, tmax, 200)
        ys = [self(t) for t in ts]
        plt.plot(ts, ys, "-", **kwargs)


class DiscreteValued:

    def get_states(self):
        if not hasattr(self, "states"):
            raise NameError("States not defined for "
                            "function.")
        return self.states

    def get_interarrival_times(self):
        if not hasattr(self, "interarrival_times"):
            raise NameError("Interarrival times not "
                            "defined for function.")
        return self.interarrival_times

    def get_arrival_times(self):
        if not hasattr(self, "interarrival_times"):
            raise NameError("Interarrival times not "
                            "defined for function.")
        return self.interarrival_times.cumsum()


def join(result1, result2):
    """Joins two result objects into a single result object.

    Args:
      result1: The first result.
      result2: The second result.
    """

    a = tuple(result1.values) if type(result1) == Tuple else (result1, )
    b = tuple(result2.values) if type(result2) == Tuple else (result2, )

    return Tuple(a + b)


def concat(*args):
    """Concatenates scalars and vectors into one data structure.

    Args:
      *args: Any number of scalar or vector objects. The last
          argument can be an InfiniteTuple.

    Returns:
      A Vector or an InfiniteTuple, depending on whether the
      last argument is an InfiniteTuple.
    """
    values = []
    for i, arg in enumerate(args):
        if is_scalar(arg):
            values.append(arg)
        elif is_vector(arg):
            values.extend(arg)
        elif isinstance(arg, InfiniteTuple):
            # check that InfiniteTuple is the last arg
            if i == len(args) - 1:
                # define concatenated InfiniteTuple
                def _func(n):
                    if n < len(values):
                        return values[n]
                    else:
                        return arg[n - len(values)]
                return type(arg)(_func)

            raise Exception("InfiniteTuple must be the last "
                            "argument to concat().")
        else:
            raise TypeError("Every argument to concat() must be either "
                            "a scalar, a vector, or an InfiniteTuple.")

    return Vector(values)


def is_scalar(x):
    return isinstance(x, (numbers.Number, str))


def is_vector(x):
    return hasattr(x, "__len__")


def is_number(x):
    return isinstance(x, numbers.Number)


def is_numeric_vector(x):
    return hasattr(x, "__len__") and all(is_number(i) for i in x)



from .base import Arithmetic, Transformable, Comparable
from .probability_space import Event
from .result import Vector, join, is_scalar, is_numeric_vector
from .results import RVResults

class RV(Arithmetic, Transformable, Comparable):
    """Defines a random variable.

    A random variable is a function which maps an outcome of
    a probability space to a number.  Simulating a random
    variable is a two-step process: first, a draw is taken
    from the underlying probability space; then, the function
    is applied to that draw to obtain the realized value of
    the random variable.

    Args:
      prob_space (ProbabilitySpace): the underlying probability space
        of the random variable.
      func (function, optional): a function that maps draws from the
        probability space to numbers. (By default, the function is the
        identity function. For named distributions, a draw from the
        underlying probability space is the value of the random
        variable itself, which is why the identity function is the
        most frequently used.)

    Attributes:
      prob_space (ProbabilitySpace): the underlying probability space
        of the random variable.
      func (function): a function that maps draws from the probability
        space to numbers.

    Examples:
      # a single draw is a sequence of 0s and 1s, e.g., (0, 0, 1, 0, 1)
      P = BoxModel([0, 1], size=5)
      # X counts the number of 1s in the draw, e.g., 5
      X = RV(P, sum)

      # the function is the identity, so Y has a Normal(0, 1) distribution
      Y = RV(Normal(0, 1))

      # a single draw from BivariateNormal is a tuple of two numbers
      P = BivariateNormal()
      # Z is the smaller of the two numbers
      Z = RV(P, min)
    """

    def __init__(self, prob_space, func=lambda x: x):
        self.prob_space = prob_space
        self.func = func

    def draw(self):
        """A function that takes no arguments and returns a single
          realization of the random variable.

        Example:
          X = RV(Normal(0, 1))
          X.draw() might return -0.9, for example.
        """
        return self.func(self.prob_space.draw())

    def sim(self, n):
        """Simulate n draws from probability space described by the random
          variable.

        Args:
          n (int): How many draws to make.

        Returns:
          RVResults: A list-like object containing the simulation results.
        """

        return RVResults(self.draw() for _ in range(n))

    def __call__(self, outcome):
        print("Warning: Calling an RV as a function simply applies the "
              "function that defines the RV to the input, regardless of "
              "whether that input is a possible outcome in the underlying "
              "probability space.")
        return self.func(outcome)

    def check_same_prob_space(self, other):
        if hasattr(other, "prob_space"):
            self.prob_space.check_same(other.prob_space)

    def apply(self, func):
        """Transform a random variable by a function.

        Args:
          func: function to apply to the random variable

        Example:
          X = RV(Exponential(1))
          Y = X.apply(log)

        Note: For most standard functions, you can apply the function to
          the random variable directly. For example, in the example above,
          Y = log(X) would have been equivalent and more readable.

        User defined functions can also be applied.

        Example:
          def g(x):
            return log(x ** 2)
          Y = X.apply(g)
        """
        def _func(outcome):
            return func(self.func(outcome))
        return RV(self.prob_space, _func)

    # This allows us to unpack a random vector,
    # e.g., X, Y = RV(BoxModel([0, 1], size=2))
    def __iter__(self):
        test = self.draw()
        if hasattr(test, "__iter__"):
            for i in range(len(test)):
                yield self[i]
        else:
            raise Exception(
                "To unpack a random vector, the RV needs to "
                "have multiple components."
            )

    def __getitem__(self, n):
        # if n is an RV, return a new random variable
        if isinstance(n, RV):
            return RV(self.prob_space,
                      lambda x: self.func(x)[n.func(x)])
        # if the indices are a list, return a random vector
        elif is_numeric_vector(n):
            return self.apply(
                lambda x: Vector(x[i] for i in n)
            )
        # if the indices are a slice, return a random vector
        elif isinstance(n, slice):
            return self.apply(
                lambda x: Vector(x[i] for i in
                                 range(n.start, n.stop, n.step or 1))
                )
        # otherwise, return the nth value
        return self.apply(lambda x: x[n])

    # The Arithmetic superclass will use this to define all of the
    # usual arithmetic operations (e.g., +, -, *, /, **, ^, etc.)
    def _operation_factory(self, op):

        def _op_func(self, other):
            # operations between this RV and another RV
            if isinstance(other, RV):
                self.check_same_prob_space(other)
                def _func(outcome):
                    return op(self.func(outcome), other.func(outcome))
                return RV(self.prob_space, _func)
            # operations between this RV and a scalar
            return self.apply(lambda x: op(x, other))

        return _op_func

    # The Comparison superclass will use this to define all of the
    # usual comparison operations (e.g., <, >, ==, !=, etc.).
    # Note that a comparison of a random variable returns an Event.
    def _comparison_factory(self, op):

        def _op_func(self, other):
            if is_scalar(other):
                return Event(self.prob_space,
                             lambda x: op(self.func(x), other))
            elif isinstance(other, RV):
                self.check_same_prob_space(other)
                return Event(self.prob_space,
                             lambda x: op(self.func(x), other.func(x)))
            raise NotImplementedError(
                "Comparisons are only defined between two RVs or "
                "between an RV and a scalar."
            )

        return _op_func


    # Define a joint distribution of two random variables: e.g., X & Y
    def __and__(self, other):
        self.check_same_prob_space(other)
        if isinstance(other, RV):
            def _func(outcome):
                return join(self.func(outcome), other.func(outcome))
        elif is_scalar(other):
            def _func(outcome):
                return join(self.func(outcome), other)
        else:
            raise Exception("Joint distributions are only defined for RVs.")
        return RV(self.prob_space, _func)

    def __rand__(self, other):
        self.check_same_prob_space(other)
        if is_scalar(other):
            def _func(outcome):
                return join(other, self.func(outcome))
        return RV(self.prob_space, _func)

    # Define conditional distribution of random variable.
    # e.g., X | (X > 3)
    def __or__(self, condition_event):
        # Check that the random variable and event are
        # defined on the same probability space.
        self.check_same_prob_space(condition_event)
        if isinstance(condition_event, Event):
            return RVConditional(self, condition_event)
        else:
            raise NotImplementedError


class RVConditional(RV):
    """Defines a random variable conditional on an event.

    RVConditionals are typically produced when you condition a
    RV on an Event object.

    Args:
      random_variable (RV): the random variable whose conditional
        distribution is desired
      condition_event (Event): the event to condition on

    Attributes:
      random_variable (RV): the random variable whose conditional
        distribution is desired
      condition_event (Event): the event to condition on

    Examples:
      X, Y = RV(Binomial(10, 0.4) ** 2)
      (X | (X + Y == 5)).draw() # returns a value between 0 and 5.
    """

    def __init__(self, random_variable, condition_event):
        self.condition_event = condition_event
        super().__init__(random_variable.prob_space,
                         random_variable.func)

    def draw(self):
        """A function that takes no arguments and returns a value from
          the conditional distribution of the random variable.

        Example:
          X, Y = RV(Binomial(10, 0.4) ** 2)
          (X | (X + Y == 5)).draw() might return a value of 4, for example.
        """
        while True:
            outcome = self.prob_space.draw()
            if self.condition_event.func(outcome):
                return self.func(outcome)



import numpy as np

from .index_sets import (
    DiscreteTimeSequence,
    Reals
)
from .probability_space import ProbabilitySpace
from .result import (
    DiscreteTimeFunction,
    ContinuousTimeFunction,
    Vector,
    is_number,
    is_numeric_vector
)
from .random_variables import RV
from .random_processes import RandomProcess

MACHINE_EPS = 1e-12


def get_gaussian_process_result(mean_func, cov_func, index_set=Reals()):

    # Determine whether the process is discrete-time or continous-time
    if isinstance(index_set, DiscreteTimeSequence):
        base_class = DiscreteTimeFunction
    elif isinstance(index_set, Reals):
        base_class = ContinuousTimeFunction
    else:
        raise Exception(
            "Index set for Gaussian process must be Reals or "
            "DiscreteTimeSequence."
        )

    class GaussianProcessResult(base_class):

        def __init__(self, mean_func, cov_func):

            self.mean = np.empty(shape=0)
            self.cov = np.empty(shape=(0, 0))
            self.observed = {}

            def _vfunc(ts):
                # This function assumes that t is an array of times.
                ts = list(ts)

                # Get current times
                times = list(self.observed.keys())

                # If this is a discrete process, t will be an index.
                # Convert it to a time.
                if isinstance(index_set, DiscreteTimeSequence):
                    ts = [t / index_set.fs for t in ts]

                # Check that every t is in the index set
                for t in ts:
                    if t not in index_set:
                        raise KeyError(
                            "Gaussian process is not defined at time %.2f." % t0
                        )

                # Create an object to store the results
                n = len(ts)
                values = np.empty(shape=n)
                values[:] = np.nan

                # Handle times that have already been calculated,
                # as well as times where the variance is 0
                i_delete = []
                for i, t in enumerate(ts):
                    if cov_func(t, t) == 0:
                        values[i] = mean_func(t)
                        i_delete.append(i)
                    elif t in self.observed:
                        values[i] = self.observed[t]
                        i_delete.append(i)
                ts = [t for i, t in enumerate(ts) if i not in i_delete]
                if not ts:
                    return values

                # Simulate values for the remaining times
                mean2 = np.array([mean_func(t) for t in ts])
                cov11 = self.cov + MACHINE_EPS * np.identity(len(times))
                cov12 = np.empty(shape=(len(times), len(ts)))
                for i, s in enumerate(times):
                    for j, t in enumerate(ts):
                        cov12[i, j] = cov_func(s, t)
                cov22 = np.empty(shape=(len(ts), len(ts)))
                for i, s in enumerate(ts):
                    for j, t in enumerate(ts):
                        cov22[i, j] = cov_func(s, t)

                cond_mean = (mean2 + (
                    cov12.T @
                    np.linalg.solve(cov11, list(self.observed.values()) - self.mean)
                ))
                cond_var = (cov22 - (
                    cov12.T @
                    np.linalg.solve(cov11, cov12)
                ))

                # update mean vector and covariance matrix
                self.mean = np.append(self.mean, mean2)
                self.cov = np.block([[cov11, cov12], [cov12.T, cov22]])

                # simulate normal with given mean and variance
                new_values = np.random.multivariate_normal(cond_mean, cond_var)

                # store the new values
                for t, v in zip(ts, new_values):
                    self.observed[t] = v
                values[np.isnan(values)] = new_values
                
                return values

            self.vfunc = _vfunc

            def _func(t):
                return _vfunc([t])[0]
            
            super().__init__(func=_func)
            self.index_set = index_set

    return GaussianProcessResult(mean_func, cov_func)


class GaussianProcessProbabilitySpace(ProbabilitySpace):

    def __init__(self, mean_func, cov_func, index_set=Reals()):
        """Initialize probability space for a Gaussian process.

        Args:
          mean_func: mean function (function of one argument)
          cov_func: (auto)covariance function (function of two arguments)
          index_set: index set for the Gaussian process
                     (by default, all real numbers)
        """

        def draw():
            return get_gaussian_process_result(
                mean_func,
                cov_func,
                index_set)

        super().__init__(draw)


class GaussianProcess(RandomProcess, RV):

    def __init__(self, mean_func, cov_func, index_set=Reals()):
        """Initialize Gaussian process.

        Args:
          mean_func: mean function (function of one argument)
          cov_func: (auto)covariance function (function of two arguments)
          index_set: index set for the Gaussian process
                     (by default, all real numbers)
        """

        prob_space = GaussianProcessProbabilitySpace(mean_func,
                                                     cov_func,
                                                     index_set)
        RandomProcess.__init__(self, prob_space)
        RV.__init__(self, prob_space)


# Define convenience class for Brownian motion
class BrownianMotionProbabilitySpace(GaussianProcessProbabilitySpace):

    def __init__(self, drift=0, scale=1):
        """Initialize probability space for Brownian motion.

        Args:
          drift: drift parameter of Brownian motion
          scale: scale parameter of Brownian motion
        """
        super().__init__(
            mean_func=lambda t: drift * t,
            cov_func=lambda s, t: (scale ** 2) * min(s, t)
        )


class BrownianMotion(RandomProcess, RV):

    def __init__(self, drift=0, scale=1):
        """Initialize Brownian motion.

        Args:
          drift: drift parameter of Brownian motion
          scale: scale parameter of Brownian motion
        """
        prob_space = BrownianMotionProbabilitySpace(
            drift=drift, scale=scale
        )
        RandomProcess.__init__(self, prob_space)
        RV.__init__(self, prob_space)



from .probability_space import ProbabilitySpace
from .random_variables import RV

def AssumeIndependent(*args):
    """Make RVs independent.

    Args:
      *args: Any number of RVs

    Returns:
      RVs with the same marginal distributions
      as the inputs, but defined on a common
      probability space so as to be independent.
    """

    # Check that none of the RVs are defined on
    # the same probability space.
    for i in range(len(args)):
        if not isinstance(args[i], RV):
            raise Exception(
                "AssumeIndependent(...) can only be "
                "used with RVs, but you passed in a "
                "%s." % type(args[i]).__name__)
        for j in range(i + 1, len(args)):
            if args[i].prob_space == args[j].prob_space:
                raise Exception(
                    "AssumeIndependent(...) can only be "
                    "called on RVs that are initially "
                    "defined on different probability "
                    "spaces."
                    )

    def draw():
        outcome = []
        for arg in args:
            outcome.append(arg.prob_space.draw())
        return outcome
    P = ProbabilitySpace(draw)

    outputs = []
    for i, arg in enumerate(args):
        # i=i forces Python to bind i now
        def _func(x, func=arg.func, i=i):
            return func(x[i])
        outputs.append(RV(P, _func))

    return tuple(outputs)



from setuptools import setup, find_packages

setup(
    name="symbulate",
    version="0.5.6",

    description="A symbolic algebra for specifying simulations.",

    url="https://github.com/dlsun/symbulate",

    author="Dennis Sun",
    author_email="dsun09@calpoly.edu",

    license="GPLv3",

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],
    
    keywords='probability simulation',

    packages=find_packages(),

    install_requires=[
        'numpy',
        'scipy',
        'matplotlib'
    ]
)



import numbers


class IndexSet(object):

    def __init__(self):
        return

    def __getitem__(self, t):
        if t in self:
            return t
        else:
            raise KeyError("Time %.2f not in index set." % t)

    def __contains__(self, value):
        return False

    def __eq__(self, other):
        return type(other) == type(self)


class Reals(IndexSet):

    def __init__(self):
        return

    def __contains__(self, value):
        try:
            return -float("inf") < value < float("inf")
        except:
            return False


class Naturals(IndexSet):

    def __init__(self):
        return

    def __contains__(self, value):
        try:
            return (
                value >= 0 and
                (isinstance(value, numbers.Integral) or
                 value.is_integer())
            )
        except:
            return False


class DiscreteTimeSequence(IndexSet):

    def __init__(self, fs):
        self.fs = fs

    def __getitem__(self, n):
        return n / self.fs

    def __contains__(self, value):
        return float(value * self.fs).is_integer()

    def __eq__(self, index):
        return (
            isinstance(index, DiscreteTimeSequence) and
            (self.fs == index.fs)
        )

class Integers(DiscreteTimeSequence):

    def __init__(self):
        self.fs = 1



import numbers
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt

from .probability_space import ProbabilitySpace
from .plot import get_next_color
from .result import Scalar, Vector, InfiniteVector

class Distribution(ProbabilitySpace):
    def __init__(self, params, scipy, discrete=True):
        self.params = params

        self.discrete = discrete

        if discrete:
            self.pmf = lambda x: scipy.pmf(x, **self.params)
            self.pdf = self.pmf # add pdf as an alias for pmf
        else:
            self.pdf = lambda x: scipy.pdf(x, **self.params)

        self.cdf = lambda x: scipy.cdf(x, **self.params)
        self.quantile = lambda x: scipy.ppf(x, **self.params)

        self.median = lambda: scipy.median(**self.params)
        self.mean = lambda: scipy.mean(**self.params)
        self.var = lambda: scipy.var(**self.params)
        self.sd = lambda: scipy.std(**self.params)
        self.sim_func = scipy.rvs

        self.xlim = (
            scipy.ppf(0.001, **self.params),
            scipy.ppf(0.999, **self.params)
            )

    def draw(self):
        return Scalar(self.sim_func(**self.params))

    # Override the inherited __pow__ function to take advantage
    # of vectorized simulations.
    def __pow__(self, exponent):
        if exponent == float("inf"):
            def draw():
                def _func(_):
                    return self.sim_func(**self.params)
                return InfiniteVector(_func)
        else:
            def draw():
                return Vector(self.sim_func(**self.params, size=exponent))
        return ProbabilitySpace(draw)

    def plot(self, xlim=None, alpha=None, ax=None, **kwargs):

        # use distribution defaults for xlim if none set
        if xlim is None:
            xlim = self.xlim

        # get the x and y values
        if self.discrete:
            xs = np.arange(int(xlim[0]), int(xlim[1]) + 1)
        else:
            xs = np.linspace(xlim[0], xlim[1], 200)
        ys = self.pdf(xs)

        # determine limits for y-axes based on y values
        ymin, ymax = ys[np.isfinite(ys)].min(), ys[np.isfinite(ys)].max()
        ylim = min(0, ymin - 0.05 * (ymax - ymin)), 1.05 * ymax
        
        # get the current axis if they exist and no axis is specified
        fig = plt.gcf()
        if ax is None and fig.axes:
            ax = plt.gca()

        # if axis already exists, set it to the union of existing and current axis
        if ax is not None:
            xlower, xupper = ax.get_xlim()
            xlim = min(xlim[0], xlower), max(xlim[1], xupper)
            ylower, yupper = ax.get_ylim()
            ylim = min(ylim[0], ylower), max(ylim[1], yupper)
        else:
            ax = plt.gca() # creates new axis

        # set the axis limits
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)
        
        # get next color in cycle
        color = get_next_color(ax)

        # plot points for discrete distributions
        if self.discrete:
            ax.scatter(xs, ys, s=40, color=color, alpha=alpha, **kwargs)

        # plot curve
        ax.plot(xs, ys, color=color, alpha=alpha, **kwargs)

        # adjust the axes, base x-axis at 0
        ax.spines["bottom"].set_position("zero")


## Discrete Distributions

class Bernoulli(Distribution):
    """Defines a probability space for a Bernoulli
         distribution.

    Attributes:
      p (float): probability (number between 0 and 1)
        of a "success" (i.e., 1)
    """

    def __init__(self, p):
        if 0 <= p <= 1:
            self.p = p
        else:
            raise Exception("p must be between 0 and 1")

        params = {
            "p" : p
            }
        super().__init__(params, stats.bernoulli, True)
        self.xlim = (0, 1) # Bernoulli distributions are not defined for x < 0 and x > 1


class Binomial(Distribution):
    """Defines a probability space for a binomial
         distribution.

    Attributes:
      n (int): number of trials
      p (float): probability (number between 0 and 1)
        that each trial results in a "success" (i.e., 1)
    """

    def __init__(self, n, p):

        if n >= 0 and isinstance(n, numbers.Integral):
            self.n = n
        #elif n == 0:
            #raise NotImplementedError
            #TODO
        else:
            raise Exception("n must be a non-negative integer")

        if 0 <= p <= 1:
            self.p = p
        else:
            raise Exception("p must be between 0 and 1")

        params = {
            "n" : n,
            "p" : p
            }
        super().__init__(params, stats.binom, True)
        self.xlim = (0, n) # Binomial distributions are not defined for x < 0 and x > n


class Hypergeometric(Distribution):
    """Defines a probability space for a hypergeometric
         distribution (which represents the number of
         ones in n draws without replacement from a box
         containing zeros and ones).

    Attributes:
      n (int): number of draws (without replacement)
        from the box
      N0 (int): number of 0s in the box
      N1 (int): number of 1s in the box
    """

    def __init__(self, n, N0, N1):

        if n > 0 and isinstance(n, numbers.Integral):
            self.n = n
        else:
            raise Exception("n must be a positive integer")

        if N0 >= 0 and isinstance(N0, numbers.Integral):
            self.N0 = N0
        else:
            raise Exception("N0 must be a non-negative integer")

        if N1 >= 0 and isinstance(N1, numbers.Integral):
            self.N1 = N1
        else:
            raise Exception("N1 must be a non-negative integer")

        params = {
            "M" : N0 + N1,
            "n" : N1,
            "N" : n
            }

        if N0 + N1 < n:
            raise Exception("N0 + N1 cannot be less than the sample size n")

        super().__init__(params, stats.hypergeom, True)
        self.xlim = (0, n) # Hypergeometric distributions are not defined for x < 0 and x > n


class Geometric(Distribution):
    """Defines a probability space for a geometric
         distribution (which represents the number
         of trials until the first success), including
         the success.

    Attributes:
      p (float): probability (number between 0 and 1)
        that each trial results in a "success" (i.e., 1)
    """

    def __init__(self, p):

        if 0 < p < 1:
            self.p = p
        else:
            raise Exception("p must be between 0 and 1")

        params = {
            "p" : p
            }
        super().__init__(params, stats.geom, True)
        self.xlim = (1, self.xlim[1]) # Geometric distributions are not defined for x < 1


class NegativeBinomial(Distribution):
    """Defines a probability space for a negative
         binomial distribution (which represents the
         number of trials until r successes), including
         the r successes.

    Attributes:
      r (int): desired number of successes
      p (float): probability (number between 0 and 1)
        that each trial results in a "success" (i.e., 1)
    """

    def __init__(self, r, p):

        if 0 < r and isinstance(r, numbers.Integral):
            self.r = r
        else:
            raise Exception("r must be a positive integer")

        if 0 < p <= 1:
            self.p = p
        else:
            raise Exception("p must be between 0 and 1")

        params = {
            "n" : r,
            "p" : p,
            "loc" : r
            }
        super().__init__(params, stats.nbinom, True)
        self.xlim = (r, self.xlim[1]) # Negative Binomial distributions are not defined for x < r

    def draw(self):
        """A function that takes no arguments and
            returns a single draw from the Negative Binomial distribution."""

        # Numpy's negative binomial returns numbers in [0, inf),
        # but we want numbers in [r, inf).
        return self.r + np.random.negative_binomial(n=self.r, p=self.p)


class Pascal(Distribution):
    """Defines a probability space for a Pascal
         distribution (which represents the number
         of trials until r successes), not including
         the r successes.

    Attributes:
      r (int): desired number of successes
      p (float): probability (number between 0 and 1)
        that each trial results in a "success" (i.e., 1)
    """

    def __init__(self, r, p):

        if 0 < r and isinstance(r, numbers.Integral):
            self.r = r
        else:
            raise Exception("r must be a positive integer")

        if 0 < p <= 1:
            self.p = p
        else:
            raise Exception("p must be between 0 and 1")

        params = {
            "n" : r,
            "p" : p
            }
        super().__init__(params, stats.nbinom, True)
        self.xlim = (0, self.xlim[1]) # Pascal distributions are not defined for x < 0


class Poisson(Distribution):
    """Defines a probability space for a Poisson distribution.

    Attributes:
      lam (float): rate parameter for the Poisson distribution
    """

    def __init__(self, lam):

        if 0 < lam:
            self.lam = lam
        else:
            raise Exception("Lambda (lam) must be greater than 0")

        params = {
            "mu" : lam
            }
        super().__init__(params, stats.poisson, True)
        self.xlim = (0, self.xlim[1]) # Poisson distributions are not defined for x < 0


class DiscreteUniform(Distribution):
    """Defines a probability space for a discrete uniform distribution.

    Attributes:
      a (int): lower bound for possible values
      b (int): upper bound for possible values
    """

    def __init__(self, a=0, b=1):
        self.a = a
        self.b = b + 1

        params = {
            "low" : self.a,
            "high" : self.b
            }

        if a >= b:
            raise Exception("b cannot be less than or equal to a")

        super().__init__(params, stats.randint, True)
        self.xlim = (a, b) # Uniform distributions are not defined for x < a and x > b


## Continuous Distributions

class Uniform(Distribution):
    """Defines a probability space for a uniform distribution.

    Attributes:
      a (float): lower bound for possible values
      b (float): upper bound for possible values
    """

    def __init__(self, a=0.0, b=1.0):
        self.a = a
        self.b = b

        params = {
            "loc" : a,
            "scale" : b - a
            }

        if a > b:
            raise Exception("b cannot be less than a")

        super().__init__(params, stats.uniform, False)
        self.xlim = (a, b) # Uniform distributions are not defined for x < a and x > b


class Normal(Distribution):
    """Defines a probability space for a normal distribution.

    Attributes:
      mean (float): mean parameter of the normal distribution
      sd (float): standard deviation parameter of the normal
        distribution
      var (float): variance parameter of the normal distribution
        (if specified, var parameter will be ignored)
    """
    #TODO edit docstring for Normal Distribution

    def __init__(self, mean=0.0, sd=1.0, var=None):

        #Note: cleaner way to implement this

        if var is None:
            if sd > 0:
                self.scale = sd
            elif sd == 0:
                raise NotImplementedError
                #TODO
            else:
                raise Exception("sd cannot be less than 0")

        else:
            if var > 0:
                self.scale = np.sqrt(var)
            elif var == 0:
                raise NotImplementedError
                #TODO
            else:
                raise Exception("var cannot be less than 0")

        params = {
            "loc" : mean,
            "scale" : self.scale
            }
        super().__init__(params, stats.norm, False)


class Exponential(Distribution):
    """Defines a probability space for an exponential distribution.
       Only one of scale or rate should be set. (The scale is the
       inverse of the rate.)

    Attributes:
      scale (float): scale parameter for gamma distribution
        (often symbolized beta = 1 / lambda)
      rate (float): rate parameter for gamma distribution
        (often symbolized lambda)
    """

    def __init__(self, rate=1.0, scale=None):

        if scale is None:
            if rate > 0:
                self.rate = rate
                self.scale = scale
            else:
                raise Exception("rate must be positive")
        else:
            if scale > 0:
                self.scale = scale
            else:
                raise Exception("scale must be positive")

        params = {
            "scale" : 1. / rate if scale is None else scale
            }
        super().__init__(params, stats.expon, False)
        self.xlim = (0, self.xlim[1]) # Exponential distributions are not defined for x < 0


class Gamma(Distribution):
    """Defines a probability space for a gamma distribution.
       Only one of scale or rate should be set. (The scale is the
       inverse of the rate.)

    Attributes:
      shape (float): shape parameter for gamma distribution
        (often symbolized alpha)
      scale (float): scale parameter for gamma distribution
        (often symbolized beta = 1 / lambda)
      rate (float): rate parameter for gamma distribution
        (often symbolized lambda)
    """

    def __init__(self, shape, rate=1.0, scale=None):

        if 0 < shape:
            self.shape = shape
        else:
            raise Exception("shape parameter must be positive")

        if scale is None:
            if rate > 0:
                self.rate = rate
                self.scale = scale
            else:
                raise Exception("rate must be positive")
        else:
            if scale > 0:
                self.scale = scale
            else:
                raise Exception("scale must be positive")

        params = {
            "a" : shape,
            "scale" : 1. / rate if scale is None else scale
            }
        super().__init__(params, stats.gamma, False)
        self.xlim = (0, self.xlim[1]) # Gamma distributions are not defined for x < 0


class Beta(Distribution):
    """Defines a probability space for a beta distribution.

    Attributes:
      a (float): alpha parameter for beta distribution
      b (float): beta parameter for beta distribution
    """

    def __init__(self, a, b):

        if 0 < a:
            self.a = a
        else:
            raise Exception("a must be positive")

        if 0 < b:
            self.b = b
        else:
            raise Exception("b must be positive")

        params = {
            "a" : a,
            "b" : b
            }
        super().__init__(params, stats.beta, False)
        self.xlim = (0, 1) # Beta distributions are not defined for x < 0 and x > 1


class StudentT(Distribution):
    """Defines a probability space for Student's t distribution.

    Attributes:
      df (int): degrees of freedom
    """

    def __init__(self, df):
        if df > 0:
            self.df = df
        else:
            raise Exception("df must be greater than 0")

        params = {
            "df" : df
            }
        super().__init__(params, stats.t, False)
        if df == 1:
            self.mean = lambda: float('nan')
            self.sd = lambda: float('nan')
            self.var = lambda: float('nan')


class ChiSquare(Distribution):
    """Defines a probability space for a chi-square distribution

    Attributes:
      df (int): degrees of freedom
    """

    def __init__(self, df):
        if df > 0 and isinstance(df, numbers.Integral):
            self.df = df
        else:
            raise Exception("df must be a positive integer")

        params = {
            "df" : df
            }
        super().__init__(params, stats.chi2, False)
        self.xlim = (0, self.xlim[1]) # Chi-Square distributions are not defined for x < 0


class F(Distribution):
    """Defines a probability space for an F distribution

    Attributes:
      dfN (int): degrees of freedom in the numerator
      dfD (int): degrees of freedom in the denominator
    """

    def __init__(self, dfN, dfD):

        if dfN > 0:
            self.dfN = dfN
        else:
            raise Exception("dfN must be greater than 0")

        if dfD > 0:
            self.dfD = dfD
        else:
            raise Exception("dfD must be greater than 0")

        params = {
            "dfn" : dfN,
            "dfd" : dfD
            }
        super().__init__(params, stats.f, False)
        self.xlim = (0, self.xlim[1]) # F distributions are not defined for x < 0


class Cauchy(Distribution):
    """Defines a probability space for a Cauchy distribution

    Attributes:
      The Cauchy distribution has no parameters
    """

    def __init__(self, loc=0, scale=1):
        self.loc = loc
        self.scale = scale

        params = {
            "loc" : loc,
            "scale" : scale
        }

        super().__init__(params, stats.cauchy, False)

    def draw(self):
        return self.loc + (self.scale * np.random.standard_cauchy())


class LogNormal(Distribution):
    """Defines a probability space for a Log-Normal distribution

       If Y has a LogNormal distribution with parameters mu and sigma, then
       log(Y) has a normal distribution with mean mu and sd sigma.

    Attributes:
      mu (float): mean of the underlying normal distribution
      sigma (float): standard deviation of the underlying normal distribution
    """

    def __init__(self, mu=0.0, sigma=1.0):

        self.norm_mean = mu

        if sigma > 0:
            self.s = sigma
            self.norm_sd = sigma
        else:
            raise Exception("sigma must be greater than 0")

        params = {
            "s" : self.s,
            "scale" : np.exp(mu)
            }
        super().__init__(params, stats.lognorm, False)
        self.xlim = (0, self.xlim[1]) # Log-Normal distributions are not defined for x < 0


class Pareto(Distribution):
    """Defines a probability space for a Pareto distribution.

    Attributes:
      b (float): shape parameter of Pareto distribution
      scale (float): scale parameter of Pareto distribution
          lower bound for possible values
    """

    def __init__(self, b=1.0, scale=1.0):

        if b > 0:
            self.b = b
        else:
            raise Exception("b must be greater than 0")

        if scale > 0:
            self.scale = scale
        else:
            raise Exception("scale must be greater than 0")

        params = {
            "b" : self.b,
            "scale" : self.scale
            }
        super().__init__(params, stats.pareto, False)
        self.xlim = (scale, self.xlim[1]) # Pareto distributions are not defined for x < scale

    def draw(self):
        """A function that takes no arguments and
           returns a single draw from the Pareto distribution."""
        
        # Numpy's Pareto is Lomax distribution, or Type II Pareto
        # but we want the more standard parametrization
        return self.scale * (1 + np.random.pareto(self.b))


# class Weibull(Distribution):
#
#     def __init__(self, eta, beta= ):


class Rayleigh(Distribution):
    """Defines a probability space for a Rayleigh distribution

    Attributes:
      The Rayleigh distribution has no parameters
    """

    def __init__(self):
        params = {}
        super().__init__(params, stats.rayleigh, False)


## Multivariate Distributions

class MultivariateNormal(Distribution):
    """Defines a probability space for a multivariate normal
       distribution.

    Attributes:
      mean (1-D array_like, of length n): mean vector
      cov (2-D array_like, of shape (n, n)): covariance matrix
    """

    def __init__(self, mean, cov):
        if len(mean) != len(cov):
            raise Exception("The dimension of the mean vector" +
                            " is not compatible with the dimensions" +
                            " of the covariance matrix.")

        if len(mean) >= 1:
            self.mean = mean
        else:
            raise Exception("Mean vector and Cov matrix cannot be empty")

        if len(cov) >= 1:
            if all(len(row) == len(mean) for row in cov):
                if np.all(np.linalg.eigvals(cov) >= 0) and np.allclose(cov, np.transpose(cov)):
                    self.cov = cov
                else:
                    raise Exception("Cov matrix is not symmetric and positive semi-definite")
            else:
                raise Exception("Cov matrix is not square")
        else:
            raise Exception("Dimension of cov matrix cannot be less than 1")

        self.discrete = False
        self.pdf = lambda x: stats.multivariate_normal(x, mean, cov)

    def plot(self):
        raise Exception(
            "Plotting is not currently available for "
            "the multivariate normal distribution."
        )

    def draw(self):
        """A function that takes no arguments and
            returns a single draw from the Multivariate Normal distribution."""

        return Vector(np.random.multivariate_normal(self.mean, self.cov))

    def __pow__(self, exponent):
        if exponent == float("inf"):
            def draw():
                def _func(n):
                    return self.draw()
                return InfiniteVector(_func)
        else:
            def draw():
                return Vector(self.draw() for _ in range(exponent))
        return ProbabilitySpace(draw)



class BivariateNormal(MultivariateNormal):
    """Defines a probability space for a bivariate normal
       distribution.

    Attributes:
      mean1 (float): mean parameter of X
      mean2 (float): mean parameter of Y
      sd1 (float): standard deviation parameter of X
      sd2 (float): standard deviation parameter of Y
      corr (float): correlation between X and Y
      var1 (float): variance parameter of X
        (if specified, sd1 will be ignored)
      var2 (float): variance parameter of Y
        (if specified, sd2 will be ignored)
      cov (float): covariance between X and Y
        (if specified, corr parameter will be ignored)
    """

    def __init__(self,
                 mean1=0.0, mean2=0.0,
                 sd1=1.0, sd2=1.0, corr=0.0,
                 var1=None, var2=None, cov=None):

        if not -1 <= corr <= 1:
            raise Exception("Correlation must be "
                            "between -1 and 1.")

        self.mean = [mean1, mean2]

        if sd1 < 0:
            raise Exception("sd1 cannot be less than 0")
        if sd2 < 0:
            raise Exception("sd2 cannot be less than 0")

        if var1 is None:
            var1 = sd1 ** 2
        if var2 is None:
            var2 = sd2 ** 2
        if var1 < 0 or var2 < 0:
            raise Exception("var1 and var2 cannot be negative")
        if cov is None:
            cov = corr * np.sqrt(var1 * var2)
        self.cov = [[var1, cov], [cov, var2]]
        self.discrete = False
        self.pdf = lambda x: stats.multivariate_normal(x, self.mean, self.cov)



class Multinomial(Distribution):
    """Defines a probability space for a multinomial
       distribution.

    Attributes:
      n (int): number of trials
      p (1-D array_like): probability vector
    """

    def __init__(self, n, p):
        if n >= 0 and isinstance(n, numbers.Integral):
            self.n = n
        #elif n == 0:
            #raise NotImplementedError
            #TODO
        else:
            raise Exception("n must be a non-negative integer")

        if sum(p) == 1 and min(p) >= 0:
            self.p = p
        else:
            raise Exception("Elements of p must be non-negative" +
                            " and sum to 1.")

        self.discrete = False
        self.pdf = lambda x: stats.multinomial(x, n, p)

    def plot(self):
        raise Exception(
            "Plotting is not currently available for "
            "the Multinomial distribution."
        )

    def draw(self):
        """A function that takes no arguments and
            returns a single draw from the multinomial distribution."""

        return Vector(np.random.multinomial(self.n, self.p))

    def __pow__(self, exponent):
        if exponent == float("inf"):
            def draw():
                def _func(_):
                    return self.draw()
                return InfiniteVector(_func)
        else:
            def draw():
                return Vector(self.draw() for _ in range(exponent))

        return ProbabilitySpace(draw)



from .index_sets import Naturals
from .random_variables import RV
from .result import TimeFunction, is_scalar


class RandomProcess(RV):
    """Defines a random process.

    A random process defines a random variable X(t)
    for each time t in an collection of times
    (called an index set).

    Args:
      prob_space (ProbabilitySpace): the underlying
        probability space for the random process.
      index_set (IndexSet): the index set for the
        random process. (By default, the index set
        is the natural numbers 0, 1, 2, 3, ....)
      func: a function that takes in an outcome from
        the probability space and a time from the
        index set and returns the value of the
        random process at that time. (By default,
        func is the canonical function. That is,
        we assume that every outcome x from the
        probability space is a function of time and
        the value of the process is simply x(t).)
    """

    def __init__(self, prob_space, index_set=Naturals(),
                 func=lambda outcome, t: outcome[t]):
        self.index_set = index_set
        # This dict stores random variables at specific times.
        self.rvs = {}

        # Define the function for the RV.
        def _func(outcome):
            def x(t):
                # First, check if the time is in self.rvs.
                if t in self.rvs:
                    return self.rvs[t].func(outcome)
                return func(outcome, t)
            return TimeFunction.from_index_set(self.index_set, x)

        super().__init__(prob_space, _func)

    def __setitem__(self, t, value):
        if t not in self.index_set:
            raise KeyError(
                "Time %s is not in the index set for this "
                "random process." % str(t)
            )
        # If value is a RV, store it in self.rvs.
        if isinstance(value, RV):
            self.rvs[t] = value
        # If value is a scalar, create and store a constant random variable
        elif is_scalar(value):
            self.rvs[t] = RV(self.prob_space, lambda outcome: value)

    def __getitem__(self, t):
        # First, check if the time is in self.rvs.
        if t in self.rvs:
            return self.rvs[t]
        return super().__getitem__(t)

    def __call__(self, t):
        return RV(self.prob_space, lambda outcome: self.func(outcome)(t))
